/* SAMHAIN file system integrity testing                                   */
/* Copyright (C) 1999 Rainer Wichmann                                      */
/*                                                                         */
/*  This program is free software; you can redistribute it                 */
/*  and/or modify                                                          */
/*  it under the terms of the GNU General Public License as                */
/*  published by                                                           */
/*  the Free Software Foundation; either version 2 of the License, or      */
/*  (at your option) any later version.                                    */
/*                                                                         */
/*  This program is distributed in the hope that it will be useful,        */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/*  GNU General Public License for more details.                           */
/*                                                                         */
/*  You should have received a copy of the GNU General Public License      */
/*  along with this program; if not, write to the Free Software            */
/*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.              */

#include "config_xor.h"

#ifdef HAVE_BROKEN_INCLUDES
#define _ANSI_C_SOURCE
#define _POSIX_SOURCE
#endif


#include <string.h>
#include <stdio.h>
#include <stdlib.h>

/* Must be early on FreeBSD
 */
#include <sys/types.h>

/* must be .le. than (1020 * 64)
 * (see sh_tools.c -- put_header)
 * 65280 = (1020*64)
 * #define TRANS_BYTES 8000  V0.8
 */
#define TRANS_BYTES 65280

/* #undef  SRP_DEBUG */
/* #define SRP_DEBUG */

#ifdef HAVE_MEMORY_H
#include <memory.h>
#endif

#ifdef TM_IN_SYS_TIME
#include <sys/time.h>
#else
#include <time.h>
#endif

#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif

#ifdef  HAVE_UNISTD_H
#include <errno.h>
#include <signal.h>
#include <setjmp.h>
#include <pwd.h>
#include <grp.h>
#include <sys/stat.h>
#include <sys/resource.h>
#include <fcntl.h>
#include <sys/wait.h>
#include <unistd.h>
#endif

#ifndef FD_SET
#define NFDBITS         32
#define FD_SET(n, p)    ((p)->fds_bits[(n)/NFDBITS] |= (1 << ((n) % NFDBITS)))
#define FD_CLR(n, p)    ((p)->fds_bits[(n)/NFDBITS] &= ~(1 << ((n) % NFDBITS)))
#define FD_ISSET(n, p)  ((p)->fds_bits[(n)/NFDBITS] & (1 << ((n) % NFDBITS)))
#endif /* !FD_SET */
#ifndef FD_SETSIZE
#define FD_SETSIZE      32
#endif
#ifndef FD_ZERO
#define FD_ZERO(p)      memset((char *)(p), '\0', sizeof(*(p)))
#endif

#if defined(HAVE_MLOCK) && !defined(HAVE_BROKEN_MLOCK)
#include <sys/mman.h>
#endif


#include <netdb.h> 
#include <sys/types.h> 
#include <netinet/in.h> 
#include <sys/socket.h> 
#include <arpa/inet.h>

#include "samhain.h"
#include "sh_tiger.h"
#include "sh_utils.h"
#include "sh_unix.h"
#include "sh_forward.h"
#include "sh_srp.h"
#include "sh_fifo.h"
#include "sh_tools.h"
#include "sh_entropy.h"
#include "sh_html.h"


/* the port client will be connecting to 
 */
#ifndef SH_DEFAULT_PORT
#define SH_DEFAULT_PORT 49777    
#endif

#ifndef SH_SELECT_REPEAT
#define SH_SELECT_REPEAT 60
#endif

#ifndef SH_HEADER_SIZE
#define SH_HEADER_SIZE 7
#endif

#ifndef SH_CHALLENGE_SIZE
#define SH_CHALLENGE_SIZE 9
#endif


#ifndef SH_STANDALONE


static int StripDomain = GOOD;

int sh_forward_set_strip (char * dummy)
{
  StripDomain = BAD;
  return GOOD;
}

#include <ctype.h>

const char *sh_strip_domain (char *name)
{
  char *          first;
  static char     name_2[SH_MINIBUF+1];
  register int    i = 0;
  register char * testit;
  int             testflag;

  if (StripDomain == BAD || (first  = strchr(name, '.')) == NULL) 
    return name;
  else
    {
      memset (name_2, '\0', SH_MINIBUF+1);

      /* check whether it is in dotted number format
       * --> last part must be kept
       */
      testit = &name[0]; testflag = GOOD;
      while (*testit != '\0') 
	{
	  if (!isdigit((int) *testit) && *testit != '.')
	    {
	      testflag = BAD;
	      break;
	    } 
	  ++testit;
	}

      if (testflag == BAD)
	{
	  first = name;
	  while (i < SH_MINIBUF && *first != '.' && *first != '\0')
	    {
	      name_2[i] = *first;
	      ++first; ++i;
	    }
	  name_2[SH_MINIBUF] = '\0';
	}
      else
	{
	  i = sl_strlen(name) - 1;
	  while (name[i] != '.' && i >= 0)
	    --i;
	  if (name[i] == '.') ++i;
	  sl_strlcpy( name_2, &name[i], SH_MINIBUF +1 );
	}
	    
    }

  return name_2;
}

/* #ifndef SH_STANDALONE */
#endif

#ifndef USE_SRP_PROTOCOL


void sh_passwd (char * salt, char * password, char * nounce, char *hash)
{

  char           *combi;
  long            len;
  register int i;
  unsigned char * dez = NULL;

  if (password == NULL)
    dez = (unsigned char *) &(skey->pw[0]);
  else if (sl_strlen(password) < PW_LEN)
    {
      fprintf(stderr, _("Password has less than %d chars !\n"),
	      PW_LEN);
      _exit(EXIT_FAILURE);
    }

  if (password == NULL)
    {
      for (i = 0; i < PW_LEN; ++i)
	{ 
	  skey->vernam[i] = (*dez); 
	  ++dez; 
	}
      skey->vernam[PW_LEN] = '\0';

      sl_strlcpy (skey->vernam,
		  sh_tiger_hash(skey->vernam, TIGER_DATA, PW_LEN), KEY_LEN+1);
      skey->vernam[KEY_LEN] = '\0';
    }
  else
    {
      sl_strlcpy (skey->vernam, password, KEY_LEN+1);
    }

  len = sl_strlen(salt) + sl_strlen(skey->vernam) + 1;
  if (nounce != NULL) len += sl_strlen(nounce);
  
  /* H(s,P)
   */
  combi = SH_ALLOC(len);
  sl_strlcpy (combi, salt, len);
  sl_strlcat (combi, skey->vernam, len);
  if (nounce != NULL)
    sl_strlcat (combi, nounce, len);
  sl_strlcpy (hash, 
	      sh_tiger_hash(combi, TIGER_DATA, sl_strlen(combi)),
	      KEY_LEN+1);

  /*
    fprintf(stderr, "DD: A: <%s>\n", salt);
    fprintf(stderr, "DD: P: <%s>\n", skey->pw); 
    fprintf(stderr, "DD: V: <%s>\n", skey->vernam); 
    fprintf(stderr, "DD: C: <%s>\n", combi);
    fprintf(stderr, "DD: H: <%s>\n", hash); 
  */

  SH_FREE (combi);
  hash[KEY_LEN] = '\0';
  return;
}
#endif

#ifdef SH_WITH_CLIENT


int sh_forward_setlogserver (char * address)
{
  if (address != NULL && sl_strlen(address) <= 256) 
    {
      sl_strlcpy (sh.logserver, address, 256+1);
      return (0);
    }
  return (-1);
}



typedef struct {
  unsigned char head[SH_HEADER_SIZE];   /* protocol, siz_hi, siz_lo */ 
  char buf[SH_BUFSIZE+1];               /* the message to transmit  */
} sh_tcp_t;



int sh_forward_send (int socket, char protocol, char * micro, 
		     char * msgbuf, unsigned long length)
{
  unsigned long           numbytes, countbytes;
  int                     flag_err = 0;
  unsigned char           head[SH_HEADER_SIZE];


  put_header (head, protocol, &length, micro);

  
  numbytes     = SH_HEADER_SIZE;
  countbytes   = write_port (socket, 
			     (char *) head, numbytes,
			     &flag_err, 300);

  if (countbytes == numbytes && msgbuf != NULL)
    {
      numbytes     = (length);
      countbytes   = write_port (socket, 
				 msgbuf, numbytes,
				 &flag_err, 300);
    }

  if (countbytes == numbytes)
    return 0;
  else
    return flag_err;
}

/* receive answer, add a trailing NULL to terminate string
 * rev 0.8
 */
long sh_forward_receive(int socket, char protocol, char * micro,          
			char * msgbuf, unsigned long length)
{
  unsigned long numbytes, countbytes;
  int           flag_err = 0;
  unsigned char head[SH_HEADER_SIZE];

  if (micro != NULL)
    micro[4]     = '\0';

  if (msgbuf != NULL)
    msgbuf[0]     = '\0';

  numbytes     = SH_HEADER_SIZE;
  countbytes   = read_port (socket, 
			    (char *) head, numbytes,
			    &flag_err, 300);

  if (countbytes != numbytes)
    {
      return flag_err;
    }
  else if (head[0] != protocol &&  (head[0] & SH_PROTO_SRP) == 0 /* not set */)
    {
      sh_error_handle(SH_ERR_ERR,
		      _("TCP: sh_forward_receive: Protocol mismatch."));
      return (-1);
    }
  else
    {
      get_header (head, &numbytes, micro);

      if (numbytes > 0)
	{
	  numbytes = (numbytes > length ? length : numbytes);

	  countbytes   = read_port (socket, 
				    msgbuf, numbytes,
				    &flag_err, 300);

	  if (countbytes < length)
	    msgbuf[countbytes] = '\0';
	  else
	    msgbuf[length-1] = '\0';

	  if (flag_err != 0)
	    return (-1);
	}
    }

  if (countbytes == numbytes)
    return numbytes;
  else
    return flag_err;
}

/**************************************************
 *
 *
 *  C L I E N T  
 *
 *
 ***************************************************/


#include <time.h>

static SH_FIFO * fifo = NULL;

static
long sh_forward_try (char * errmsg);


long sh_forward (char * errmsg)
{
  static int           have_server = GOOD;
  int    status;
  char * popmsg;

  /* No log server available.
   */
  if (have_server == GOOD && sh.logserver[0] == '\0')
    {
      have_server = BAD;
      sh_error_handle(SH_ERR_ERR, _("No log server given."));
      return (-1);
    }
  else if (have_server == BAD)
    {
      return (-1);
    }

  if (fifo == NULL)
    {
      fifo = SH_ALLOC(sizeof(SH_FIFO));
      fifo_init(fifo);
    }

  /* Check for messages on the queue, and send them first
   */
  while (NULL != (popmsg = pop_list(fifo)) )
    {
      status = sh_forward_try (popmsg);
      if (status != 0)
	{
	  push_tail_list (fifo, popmsg);    
	  push_list (fifo, errmsg);   
	  SH_FREE(popmsg);
	  return (-1);
	}
      SH_FREE(popmsg);
    }

  /* Now send the error message.
   */ 
  status = sh_forward_try (errmsg);
  if (status != 0)
    {
      push_list (fifo, errmsg);    /* push message on stack */
      return (-1);
    }    

  return 0;  
}

static 
long sh_forward_try_impl (char * errmsg, char what);

long sh_forward_try (char * errmsg)
{
  return (sh_forward_try_impl (errmsg, SH_PROTO_MSG));
}

long sh_forward_req_file (char * file)
{
  return (sh_forward_try_impl (file, SH_PROTO_BIG));
}

static  
long sh_forward_try_impl (char * errmsg, char what)
{
  static int           initialized = BAD;
  static int           conn_state  = GOOD;
  int                  sockfd;
  int                  flag_err;
  char               * answer;

  unsigned char        theProto;

  char                 hash[KEY_LEN+1];
  long                 len;
  char               * buffer;

  char   foo_u[5];

  char   nsrv[KEY_LEN+1];
  char   nclt[KEY_LEN+1];
  SL_TICKET sfd;
  int    transfercount;

  char   foo_M1[KEY_LEN+1];
  UINT32 ticks;

  char   error_msg[256];

#ifdef USE_SRP_PROTOCOL
  char   u_real[SH_CHALLENGE_SIZE];
  char * foo_A;
  char * foo_Sc;
  char * M;
#else
  char   nounce[KEY_LEN+1];
  char   temp[2*KEY_LEN+1];
  char   nonce_u[KEY_LEN+1];
#endif
  
  static unsigned long time_now  = 1200;
  static unsigned long time_last =    0;

  /* No message to transmit.
   */
  if (errmsg == NULL && initialized == GOOD)
    return 0;
  
  /* Connection in bad state.
   */
  if (initialized == BAD || conn_state == BAD)
    {
      /* retry bad attempt only after some time (10 minutes)
       */
      time_now  = (unsigned long) time (NULL);
      if ((time_now - time_last) < 600) 
	{
	  return (-1);
	}
    }
  time_last  = (unsigned long) time (NULL);


  /* try to connect to log server
   */
  sockfd = connect_port (sh.logserver, SH_DEFAULT_PORT, error_msg, 256);

  if (sockfd < 0)
    {
      conn_state = BAD;
      sh_error_handle(SH_ERR_FATAL,
		      error_msg, sh.logserver);
      return (-1);
    }

  conn_state = GOOD;

  /*************************
   *
   *  initialization
   * 
   */

  flag_err = 0;
  answer   = SH_ALLOC(512);
  MLOCK(answer, 512);

#ifndef USE_SRP_PROTOCOL

  /* challenge/response authentication
   */
  if (initialized == BAD)
    {
      theProto = SH_PROTO_SRP;

#ifdef SRP_DEBUG
      fprintf(stderr, "entry\n");
#endif

      sl_strlcpy (answer, sh.localhost, 512);

      flag_err = sh_forward_send (sockfd, theProto, _("SALT"), 
				  answer,  sl_strlen(answer));

#ifdef SRP_DEBUG
      fprintf(stderr, "Sent: SALT,  flag_err = %d\n", flag_err);
#endif

      /* get nonce from server
       */
      if (flag_err == 0)
	{
	  flag_err = sh_forward_receive (sockfd, theProto, foo_u, 
					 answer,  511);
	  flag_err = (flag_err < 0) ? flag_err : 0;
	}

#ifdef SRP_DEBUG
      fprintf(stderr, "Recv: nonce, flag_err = %d\n", flag_err);
#endif

      /* entry point for jump from message forward if session key must
       * be re-initialized
       */	 
 initBlock:

      if (0 == check_request (foo_u, _("INIT")) && 
	  flag_err == 0                         &&
	  sl_strlen(answer) >  KEY_LEN )
	sl_strlcpy(nounce, &answer[KEY_LEN], KEY_LEN+1);
      else
	flag_err = (-1);

#ifdef SRP_DEBUG
      fprintf(stderr, "Recv: INIT,  flag_err = %d\n", flag_err);
#endif

      /* verify random nonce v from server H(v, P)v
       */
      sh_passwd (nounce, NULL, NULL, temp);
      if ( 0 != sl_strncmp(temp, answer, KEY_LEN))
	flag_err = (-1);

#ifdef SRP_DEBUG
      fprintf(stderr, "Vefy: nonce, flag_err = %d\n", flag_err);
#endif

      /* create own nonce
       */
      ticks = (UINT32) taus_get (&(skey->rng0[0]), 
				 &(skey->rng1[0]),
				 &(skey->rng2[0]));
      sl_strlcpy(nonce_u, 
		 sh_tiger_hash((char *) &ticks, 
			       TIGER_DATA, sizeof(UINT32)),
		 KEY_LEN+1);

      /* form the message H(H(u,v),P)u
       */
      sl_strlcpy(temp, nonce_u, 2*KEY_LEN+1); 
      sl_strlcat(temp,  nounce, 2*KEY_LEN+1); 
      sl_strlcpy(temp, 
		 sh_tiger_hash(temp, 
			       TIGER_DATA, sl_strlen(temp)),
		 KEY_LEN+1);
      sh_passwd (temp, NULL, NULL, foo_M1);
      sl_strlcpy(temp, foo_M1, 2*KEY_LEN+1);
      sl_strlcat(temp, nonce_u, 2*KEY_LEN+1);

      /* send it to server
       */
      if (flag_err == 0)
	flag_err = sh_forward_send (sockfd, (theProto|SH_PROTO_SRP), 
				    _("PASS"), temp, 
				    sl_strlen(temp));

#ifdef SRP_DEBUG
      fprintf(stderr, "Sent: PASS,  flag_err = %d\n", flag_err);
#endif

      if (flag_err == 0)
	{
	  flag_err = sh_forward_receive (sockfd, (theProto|SH_PROTO_SRP), 
					 foo_u, answer,  511);  
	  sh_passwd (nounce, NULL, nonce_u, foo_M1);
	  sl_strlcpy (skey->session, foo_M1, KEY_LEN+1);
	  initialized = GOOD;
	}
      else 
	{
	  sh_error_handle(SH_ERR_ERR,
			  _("TCP: AUTH: failed"));
	}

      if (initialized == BAD)
	{
	  sh_error_handle(SH_ERR_ERR, 
			  _("Session key negotiation failed"));
	  memset(answer, '\0', sizeof(answer));
	  MUNLOCK(answer, 512);
	  SH_FREE(answer);
	  return (-1);
	}
      else
	{
	  sh_error_handle(SH_ERR_INFO,
			  _("Session key negotiated successfully"));
	}
    }

#else

  /* use SRP protocol           */
  /* #include "srp_client.frag" */

  /* This is the SRP authenticated key exchange protocol.
   * Produces a session key skey->session.
   */
  if (initialized == BAD)
    {
#ifdef SRP_DEBUG
      fprintf(stderr, "DD Client:  : <ENTRY>\n");
#endif

      theProto = SH_PROTO_SRP;

      sl_strlcpy (answer, sh.localhost, 512);
      flag_err = sh_forward_send    (sockfd, theProto, _("SALT "), 
				     answer,  sl_strlen(answer)); 

#ifdef SRP_DEBUG
      fprintf(stderr, "Sent: SALT %s flag_err = %d\n", answer, flag_err);
#endif

      if (flag_err == 0)
	{
	  flag_err = sh_forward_receive (sockfd, theProto, foo_u, 
					 answer,  511);
	  flag_err = (flag_err < 0) ? flag_err : 0;
	}
	  
	  
#ifdef SRP_DEBUG
      fprintf(stderr, "Recv: nonce, flag_err = %d\n", flag_err);
#endif

      /* entry point for jump from message forward if session key must
       * be re-initialized
       */	 
    initBlock:
      
#ifdef SRP_DEBUG
      fprintf(stderr, "DD Client:  : <INIT ..>\n");
#endif

      if (flag_err == 0 &&
	  (0 == check_request (foo_u, _("INIT"))))
	{
	  if (0 != sh_srp_init())
	    sh_error_handle(SH_ERR_ERR,
			    _("Error initializing bignum library."));
	  else /* if (0 == sh_srp_init()) */
	    {
#ifdef SRP_DEBUG
	      fprintf(stderr, "DD Client:  : <... SUCCESS>\n");
#endif
	      sh_srp_x (answer, NULL);  /* x        password      */
	      sh_srp_make_a ();         /* a        random number */
	      foo_A = sh_srp_A();       /* g^a                    */

#ifdef SRP_DEBUG
	      fprintf(stderr, "DD Client: A: <%s>\n", foo_A);
#endif

	      if (foo_A == NULL)
		flag_err = (-1);

	      if (flag_err == 0)
		flag_err = sh_forward_send    (sockfd, 
					       (theProto|SH_PROTO_SRP), 
					       _("PC01"),
					       foo_A, sl_strlen(foo_A)+1); 
	      if (flag_err == 0)
		{
		  flag_err = sh_forward_receive (sockfd, 
						 (theProto|SH_PROTO_SRP),
						 foo_u,
						 answer, 511);
		  flag_err = (flag_err < 0) ? flag_err : 0;
		}
	  
#ifdef SRP_DEBUG
	      fprintf(stderr, "DD Client: B: <%s>\n", answer);
	      fprintf(stderr, "DD Client: u: <%s>\n", foo_u);
#endif

	      /*                     u        nounce        */
	      /*                     B        answer        */
	      /*                     S = (B-g^x)^(a+ux)     */
	      
	      if (flag_err == 0)
		{ 
		  if (0 != sh_srp_check_zero (answer))
		    sh_error_handle(SH_ERR_ERR,
				    _("SRP: B == 0 ."));
		  else /* if (0 != sh_srp_check_zero (answer)) */
		    {
		      sl_strlcpy(u_real, sh_tiger_hash(foo_u, TIGER_DATA, 4), 
				 SH_CHALLENGE_SIZE);
		      foo_Sc = sh_srp_S_c (u_real, answer);
#ifdef SRP_DEBUG
		      fprintf(stderr, "DD Client: U: <%s>\n", u_real);
		      fprintf(stderr, "DD Client:Sc: <%s>\n", foo_Sc);
		      fprintf(stderr, "DD Client: K: <%s>\n", 
			      sh_tiger_hash(foo_Sc, 
					    TIGER_DATA, 
					    sl_strlen(foo_Sc)));
#endif
		      /* now send H(A,B,H(Sc)) and check 
		       */
		      if (foo_Sc != NULL)
			{
			  sl_strlcpy (foo_M1, 
				      sh_srp_M (foo_A, 
						answer, 
						sh_tiger_hash(foo_Sc, 
							      TIGER_DATA, 
							      sl_strlen(foo_Sc))),
				      KEY_LEN+1);
#ifdef SRP_DEBUG
			  fprintf(stderr, "DD Client:M1: <%s>\n", foo_M1);
#endif
			  flag_err = sh_forward_send(sockfd, 
						     (theProto|SH_PROTO_SRP), 
						     _("PC02"),
						     foo_M1, KEY_LEN+1);
			}
		      else
			{
			  flag_err = (-1);
			}

		      if (flag_err == 0)
			{
			  flag_err =sh_forward_receive(sockfd, 
						       (theProto|SH_PROTO_SRP),
						       foo_u, 
						       answer, 511);
			  flag_err = (flag_err < 0) ? flag_err : 0;
			}

#ifdef SRP_DEBUG
		      fprintf(stderr, "DD Client:M : <%s>\n", answer);
#endif
		      if (flag_err == 0   &&
			  (0 == check_request (foo_u, _("PARP"))))
			{
			  /* ------  verify M2 = H(A, M1, K) --------
			   */
			  M = sh_srp_M (foo_A, foo_M1,
					sh_tiger_hash(foo_Sc,
						      TIGER_DATA,
						      sl_strlen(foo_Sc)));
			  if (M != NULL && 0 == sl_strncmp (answer, M, KEY_LEN+1))
			    {
			      sl_strlcpy (skey->session, 
					  sh_tiger_hash(foo_Sc, 
							TIGER_DATA,
							sl_strlen(foo_Sc)),
					  KEY_LEN+1);
			      initialized = GOOD;
			    }
			  else
			    {
			      sh_error_handle(SH_ERR_ERR,
					      _("SRP: Session key mismatch."));
			    }
			}
		      if (foo_Sc != NULL)
			SH_FREE(foo_Sc);
		    }
		}
	      if (foo_A != NULL)
		SH_FREE(foo_A);
	      sh_srp_exit();
	    }
	}

      if (initialized == BAD)
	{
	  sh_error_handle(SH_ERR_ERR,
			  _("Session key negotiation failed"));
	  memset(answer, '\0', sizeof(answer));
	  MUNLOCK(answer, 512);
	  SH_FREE(answer);
	  return (-1);
	}
      else
	{
	  sh_error_handle(SH_ERR_INFO,
			  _("Session key negotiated successfully"));
	}
    }

#endif

  /* no message, just session key negotiated
   */
  if (errmsg == NULL)
    {
      memset(answer, '\0', sizeof(answer));
      MUNLOCK(answer, 512);
      SH_FREE(answer);
      return (0);
    }
  else if (what == SH_PROTO_BIG)
    {
      SH_FREE (answer);
      answer   = SH_ALLOC(TRANS_BYTES + 50);
      MLOCK(answer, TRANS_BYTES + 50);
    }

  sl_strlcpy (answer, sh.localhost, 512);
  sl_strlcat (answer, skey->session, 512);
  sl_strlcpy (answer, sh_tiger_hash(answer, TIGER_DATA, sl_strlen(answer)), 
	      KEY_LEN+1);
  sl_strlcat (answer, sh.localhost, 512);

  /***********************************************
   *
   * send the message
   *
   */
  if (what == SH_PROTO_MSG)
    {
      theProto = SH_PROTO_MSG;

      /* say HELO       
       */

      flag_err = sh_forward_send    (sockfd, theProto, _("HELO"),
				     answer, sl_strlen(answer)); 
      if (flag_err == 0)
	{ 
	  /* get challenge  
	   */
	  flag_err = sh_forward_receive (sockfd, SH_PROTO_MSG, foo_u, 
					 answer, 255);   
	  flag_err = (flag_err < 0) ? flag_err : 0;
	  
	  if (flag_err == 0)
	    {
	      
	      /* re-negotiate key
	       */
	      if (0 == check_request_nerr(foo_u, _("INIT")))
		{
		  flag_err    = 0;
		  initialized = BAD;
		  goto initBlock;
		}
	      
	      else if (0 == check_request(foo_u, _("TALK")))
		{
		  /* save challenge  
		   */
		  sl_strlcpy(nsrv, answer, KEY_LEN + 1);
		  
		  /* Hash(msg,challenge,sessionkey)  
		   */
		  sl_strlcat(answer, skey->session, 512);
		  if (sl_strlen(answer) > 2*KEY_LEN)
		    len    = sl_strlen(errmsg) + sl_strlen(answer) + 1;
		  else
		    len    = sl_strlen(errmsg) + 2*KEY_LEN + 1;
		  buffer = SH_ALLOC(len);
		  
		  MLOCK(buffer, len);
		  
		  sl_strlcpy(buffer, errmsg, len);
		  sl_strlcat(buffer, answer, len);
		  sl_strlcpy(hash, 
			     sh_tiger_hash (buffer, TIGER_DATA, 
					    sl_strlen(buffer)), 
			     KEY_LEN+1);
		  sl_strlcpy(buffer, errmsg, len);
		  sl_strlcat(buffer, hash,   len);
		  flag_err = sh_forward_send    (sockfd,SH_PROTO_MSG,_("MESG"),
						 buffer, sl_strlen(buffer)+1);

		  if (flag_err == 0)
		    {
		      /* get confirmation
		       */
		      flag_err = sh_forward_receive (sockfd, 
						     SH_PROTO_MSG|SH_PROTO_END,
						     foo_u, 
						     answer, 255);   
		      flag_err = (flag_err < 0) ? flag_err : 0;
		    }
		  if (flag_err == 0)
		    {
		      sl_strlcpy(buffer, errmsg, len);
		      sl_strlcat(buffer, skey->session, len);
		      sl_strlcat(buffer, nsrv, len);

		      flag_err = sl_strncmp(answer,
					 sh_tiger_hash (buffer, TIGER_DATA,
							sl_strlen(buffer)),
					 KEY_LEN);
		      if (flag_err != 0)
			{
#ifdef ENOMSG
			  flag_err = ENOMSG;
#else
			  flag_err = EIO;
#endif
			  sh_error_handle(SH_ERR_ERR,
					  _("No message delivery "\
					    "confirmation from server."));
			}
		      else
			sh_error_handle(SH_ERR_INFO,
					  _("Message delivery "\
					    "confirmed by server."));
			  
		    }

		  memset(buffer, '\0', len);
		  MUNLOCK(buffer, len);
		  SH_FREE(buffer);
		}
	      else
		{
		  /* unexpected reply from server
		   */
		  flag_err = (-1);
		}
	    }
	}
    }


  else if (what == SH_PROTO_BIG)
    {
      theProto = SH_PROTO_BIG;

      /* say HELO       
       */
      flag_err = sh_forward_send    (sockfd, theProto, _("HELO"),
				     answer, sl_strlen(answer));

      if (flag_err == 0)
	{ 
	  /* get NSRV  
	   */
	  flag_err = sh_forward_receive (sockfd, SH_PROTO_BIG, foo_u, 
					 answer, 255);
	  flag_err = (flag_err < 0) ? flag_err : 0;
	}   
	  
      if (flag_err == 0)
	{
	  
	  /* re-negotiate key
	   */
	  if (0 == check_request_nerr(foo_u, _("INIT")))
	    {
	      flag_err    = 0;
	      initialized = BAD;
	      goto initBlock;
	    }
	  
 
	  else if (0 == check_request(foo_u, _("NSRV")))
	    {
	      sl_strlcpy(nsrv, answer, KEY_LEN+1);
	      
	      ticks = (UINT32) taus_get (&(skey->rng0[0]), 
					 &(skey->rng1[0]),
					 &(skey->rng2[0]));
              
	      sl_strlcpy(nclt, 
			 sh_tiger_hash((char *) &ticks, 
				       TIGER_DATA, sizeof(UINT32)),
			 KEY_LEN+1);
	      buffer = sh_util_strconcat (nsrv, nclt, 
					  skey->session, NULL);
	      sl_strlcpy(foo_M1, sh_tiger_hash(buffer, TIGER_DATA,
					       sl_strlen(buffer)),
			 KEY_LEN+1);
	      memset (buffer, '\0', sl_strlen(buffer));

	      sl_strlcpy(buffer, nclt, KEY_LEN+1);
	      sl_strlcat(buffer, errmsg, KEY_LEN+5);
	      
	      flag_err = sh_forward_send    (sockfd, theProto, _("NCLT"),
					     buffer, sl_strlen(buffer));
 
	      SH_FREE (buffer);
	    }   
	}

      if (flag_err == 0)
	{
	  /* receive the file
	   */

          /* open a temporary file
           */
          if ( (sfd = open_tmp ()) < 0)
	    {
	      flag_err = (-1);
	      sh_error_handle(SH_ERR_ERR,
			      _("Could not open temporary file."));

	    }
	  else
	    {
	      /* read from socket into tmp file
	       */
	      transfercount = 0;
	      flag_err      = 0;

	      do {
		flag_err = sh_forward_receive (sockfd, SH_PROTO_BIG, foo_u, 
					       answer, TRANS_BYTES + 50);

		sh_error_handle(SH_ERR_ALL,
				_("Received: %d bytes, marked %s."),
				flag_err, foo_u);
		

		if (flag_err > 0 && 0 == check_request_nerr(foo_u, _("FILE")))
		  {
		    if (0 == hash_check (foo_M1, answer, flag_err))
		      {
			sl_write(sfd, &answer[KEY_LEN], flag_err-KEY_LEN);
			++transfercount;
			flag_err = sh_forward_send (sockfd, theProto, 
						    _("RECV"),
						    nclt, sl_strlen(nclt));
		      }
		    else
		      break;
		  }
		else
		  break;
	      } while (transfercount < 32000); /* 64 Mbyte */
		    
	      if (0 == check_request_nerr(foo_u, _("EEOT")) &&
		  0 <  flag_err                             &&
		  0 == hash_check (foo_M1, answer, flag_err))
		{
		  flag_err = sh_forward_send (sockfd, theProto, _("EOTE"),
					    nclt, sl_strlen(nclt));

		  rewind_tmp (sfd);
		  sl_sync(sfd);
		  sh_error_handle(SH_ERR_INFO,
				  _("File download completed."));
		}
	      else
		{
		  sh_error_handle(SH_ERR_ERR,
				  _("File download failed."));
		  sl_close (sfd);
		  sfd = (-1);
		}

	      close (sockfd);
	      memset(answer, '\0', TRANS_BYTES + 50);
	      MUNLOCK(answer, TRANS_BYTES + 50);
	      SH_FREE(answer);
	      
	      return (sfd);
	    }
	}

      close (sockfd);
      memset(answer, '\0', TRANS_BYTES + 50);
      MUNLOCK(answer, TRANS_BYTES + 50);
      SH_FREE(answer);
      sh_error_handle(SH_ERR_ERR,
		      _("File download failed."));
      return (-1);
    }
		  
      
 
  close (sockfd);
  memset(answer, '\0', 512);
  MUNLOCK(answer, 512);
  SH_FREE(answer);

#ifdef SH_ERROR_H  
  if (flag_err != 0)
    {
      conn_state = BAD;
      sh_error_handle(SH_ERR_ERR,
		      _("Connection error: %s"),
		      sh_error_message(flag_err));
      return (-1);
    }
#endif

  return 0;
}

/* #ifdef SH_WITH_CLIENT */
#endif

#if defined (SH_WITH_CLIENT) || defined (SH_WITH_SERVER)

#if 0
void sh_forward_setpwd(char * hname, char * key)
{
  pid_t    pid;
  char prg[1024];
  char file1[1024];

  struct stat buf;

  char * arg[6];

  if (NULL != getcwd(prg, 512))
    {
      sl_strlcat(prg, _("/"), 1024);
      sl_strlcat(prg, _("samhain_setpwd"), 1024);
    }

  if (EINTR_RETRY(stat(prg, &buf)) >= 0)  /* use current directory */
    {
      sl_strlcpy(file1, _("./samhain"), 1024);
    }
  else                       /* use install directory */
    {
      sl_strlcpy(prg, SH_INSTALL_DIR, 512);
      sl_strlcat(prg, _("samhain_setpwd"), 1024);
      sl_strlcpy(file1, SH_INSTALL_DIR, 512);
      sl_strlcat(file1, _("samhain"), 1024);
    }

  arg[0] = prg;
  arg[1] = file1;
  arg[2] = hname;
  arg[3] = key;
  arg[4] = NULL;

  fprintf(stderr, "%s %s %s %s\n", prg, file1, hname, key);
  fprintf(stderr, "%s %s %s %s\n", arg[0], arg[1], arg[2], arg[3]);

  pid = fork();

  switch (pid)
    {
    case -1:                 /* failed */
      _exit(EXIT_FAILURE);
      return;
    case 0:                  /* child  */
      execve (prg, arg, NULL);
      _exit(EXIT_FAILURE);
      return;
    }

  return;               
}
#endif
 
#ifndef USE_SRP_PROTOCOL

int sh_forward_make_client (char * str)
{
  /* char *          safer; */
  char            key[KEY_LEN+1];
  unsigned char   in[PW_LEN+1];
  int    i = 0, j, k, l = 0;
  
  if (sl_strlen(str) != (PW_LEN * 2)) 
    {
      fprintf(stderr, 
              _("Input must be a %d digit hexadecimal number"\
                " (only 0-9, a-f, A-F allowed in input)\n"),
              (PW_LEN * 2));
      _exit(EXIT_FAILURE);
    }
  
  while (i < (PW_LEN * 2))
    {
      k = sh_util_hexchar(str[i]); j = sh_util_hexchar(str[i+1]); 
      if (k != -1 && j != -1) 
        {
          in[l] = (k * 16 + j);
          ++l; i+= 2;
        }
      else
        {
          fprintf(stderr, _("Invalid char %c\n"), str[i]);
          _exit(EXIT_FAILURE);
        }
    }
  in[PW_LEN] = '\0';

  
  /*
    for (i=0; i<8; ++i) 
    if (in[i] == '\0')  in[i] = '\001';
    safer = sh_util_safe_name (in);
    fprintf(stdout, _("Password |%s|\n"), safer);
    SH_FREE(safer);
  */
  
  sl_strlcpy (key, 
	      sh_tiger_hash ((char *)in, TIGER_DATA, PW_LEN), 
	      KEY_LEN+1);
  key[KEY_LEN] = '\0';
  
  fprintf(stdout, _("Client entry: Client=HOSTNAME@00000000@%s\n"), 
          key);
  fflush(stdout);

  _exit(EXIT_SUCCESS);
  return 0;
}

#else

int sh_forward_make_client (char * str)
{
  char * foo_v;
  /* char * safer; */
  char   salt[17];
  char   key[KEY_LEN+1];
  char   in[PW_LEN];
  int    i = 0, j, k, l = 0;
  
  if (sl_strlen(str) != (PW_LEN*2)) 
    {
      fprintf(stderr, 
	      _("Input must be a %d digit hexadecimal number"\
                " (only 0-9, a-f, A-F allowed in input)\n"),
	      (PW_LEN*2));
      _exit(EXIT_FAILURE);
    }

    while (i < (PW_LEN*2))
      {
        k = sh_util_hexchar(str[i]); j = sh_util_hexchar(str[i+1]); 
        if (k != -1 && j != -1) 
          {
            in[l] = (k * 16 + j);
            ++l; i+= 2;
          }
        else
          {
            fprintf(stderr, _("Invalid char %c\n"), str[i]);
            _exit(EXIT_FAILURE);
          }
      }
    
  /*
    for (i=0; i<8; ++i) 
    if (in[i] == '\0')  in[i] = '\001';
  */
  
  if (0 == sh_srp_init())
    {
      /*
	safer = sh_util_safe_name (in);
	fprintf(stdout, _("Password |%s|\n"), safer);
	SH_FREE(safer);
      */
      sh_util_keyinit(key, KEY_LEN);
      sl_strlcpy(salt, sh_tiger_hash(key, TIGER_DATA, KEY_LEN), 17); 
      sh_srp_x (salt, in);
      foo_v  = sh_srp_verifier ();
      fprintf(stdout, _("Client=HOSTNAME@%s@%s\n"), 
	     salt, foo_v);
      fflush(stdout);
      SH_FREE(foo_v);
      sh_srp_exit();
      _exit(EXIT_SUCCESS);
    }
  fprintf(stdout, _("ERROR initializing BigNum library.\n"));
  fflush (stdout);
  _exit(EXIT_FAILURE);
  return -1;
}
#endif

#endif

int sh_forward_create_password ()
{
  UINT32   val[2]; 
  char     output[KEY_LEN+1];

  val[0] = taus_get (&(skey->rng0[0]), &(skey->rng0[1]), &(skey->rng0[2]));
  val[1] = taus_get (&(skey->rng0[0]), &(skey->rng0[1]), &(skey->rng0[2]));

  sl_strlcpy (output, 
	      sh_tiger_hash((char *)(&val[0]), TIGER_DATA, 2*sizeof(UINT32)),
	      KEY_LEN);

  output[16] = '\0';

  fprintf(stdout, _("%s\n"), output);
  fflush (stdout);

  _exit(EXIT_SUCCESS);
  return (0);  /* avoid compiler warning */
}

/**************************************************
 *
 *
 *  S E R V E R   
 *
 *
 ***************************************************/

#ifdef SH_WITH_SERVER

#include "sh_readconf.h"


#define CONN_FREE    0
#define CONN_READING 1
#define CONN_SENDING 2
#define CONN_PAUSE   3
#define CONN_BUSY    4

char * clt_stat[] = {
  N_("Inactive"),
  N_("Started"),
  N_("ILLEGAL"),
  N_("FAILED"),
  N_("Exited"),
  N_("PANIC"),
  N_("POLICY"),
  N_("File transfer"),
  N_("Message"),
};



#include <time.h>

/* in sh_html.h:
 *  typedef struct client_entry {
 *  } client_t;
 */

client_t * all_clients = NULL;

void sh_forward_html_write()
{
  sh_html_write(all_clients);
  return;
}

void free_client(client_t * here)
{
  if (here == NULL)
    return;

  if (here->hostname != NULL)
    SH_FREE(here->hostname);
  if (here->salt != NULL)
    SH_FREE(here->salt);
  if (here->verifier != NULL)
    SH_FREE(here->verifier);
  SH_FREE(here);
  return;
}



int sh_forward_register_client (char * str)
{
  client_t   * newclt;
  client_t   * here;
  client_t   * up;

  char       * ptr;
  int          sepnum = 0;
  int          sep[2];
  register int i = 0;
  int          treetest;
  int          siz_str = 0;

  ptr = str; 
  while (*ptr) {
    if (*ptr == '@' && sepnum < 2 ) 
      { 
	sep[sepnum] = i;
	++sepnum;
      } 
    ++ptr; ++i; 
  }

  
  if (sepnum == 2)
    {
      newclt = SH_ALLOC (sizeof(client_t));
      newclt->hostname = SH_ALLOC (sep[0]+1);
      newclt->salt     = SH_ALLOC (sep[1]-sep[0]);
      newclt->verifier = SH_ALLOC (sl_strlen(str)-sep[1]+1);
      newclt->exit_flag         = 0;
      newclt->dead_flag         = 0;
      newclt->session_key[0]    = '\0';
      newclt->last_connect      = (time_t) 0;
      newclt->session_key_timer = (time_t) 0;
      newclt->status_now   = CLT_INACTIVE;
      for (i = 0; i < CLT_MAX; ++i) 
	newclt->status_arr[i] = CLT_INACTIVE;
      sl_strlcpy(newclt->timestamp[CLT_INACTIVE],   sh_unix_time(0), TIM_MAX);
      sl_strlcpy(newclt->hostname,  &str[0],        sep[0]+1);
      sl_strlcpy(newclt->salt,      &str[sep[0]+1], sep[1]-sep[0]);
      sl_strlcpy(newclt->verifier,  &str[sep[1]+1], sl_strlen(str)-sep[1]+1);

      /* insert in binary search tree
       */
      newclt->left     = NULL;
      newclt->right    = NULL;
      newclt->up       = NULL;
      here = all_clients;
      up   = NULL;
      while (here) 
	{
	  treetest = sl_strcmp (newclt->hostname, here->hostname);

	  /* if the entry exists already, just update the salt && verifier
	   */
	  if (treetest == 0)
	    {
	      SH_FREE(here->verifier);
	      siz_str = strlen (newclt->verifier) + 1;
	      here->verifier = SH_ALLOC (siz_str);
	      sl_strlcpy(here->verifier, newclt->verifier, siz_str);

	      SH_FREE(here->salt);
	      siz_str = strlen (newclt->salt) + 1;
	      here->salt = SH_ALLOC (siz_str);
	      sl_strlcpy(here->salt, newclt->salt, siz_str);

	      here->dead_flag = 0;
	      
	      free_client(newclt);
	      
	      newclt = here;

	      /* heaven help, it's a goto !!
	       */
	      goto registeredClient;

	    }
	  up   = here;
	  here = (treetest < 0) ? here->left : here->right;
	}
      newclt->up = up;
      
      if(up)
	{
	  if(sl_strcmp (newclt->hostname, up->hostname) < 0)
	    up->left  = newclt;
	  else
	    up->right = newclt;
	}
      else
	all_clients = newclt;


    registeredClient:

      sh_error_handle(SH_ERR_ALL,
		      _("Registered %s, salt %s, verifier %s"),
		      newclt->hostname,
		      newclt->salt,
		      newclt->verifier);
    }
  else
    {
      sh_error_handle(SH_ERR_ERR,
		      _("Bad client registry entry %s"),
		      str);
      return (-1);
    } 
      
  return (0);
}

typedef struct {
  int             state;
  int             fd;
  char          * buf;
  unsigned char   head[SH_HEADER_SIZE];
  char            challenge[SH_CHALLENGE_SIZE];
  char            peer[SH_MINIBUF+1];
  client_t      * client_entry;
  char          * K;
  char          * M1;
  char          * A;
  int             headcount;
  unsigned long   bytecount;
  unsigned long   bytes_to_send;
  unsigned long   bytes_to_get;
  int             pass;
  unsigned long   timer;

  char          * FileName;
  unsigned long   FileLength;
  unsigned long   FileSent;
  char            FileType[5];
} sh_conn_t;



void sh_forward_do_free (sh_conn_t * conn)
{
  if (conn->K != NULL) 
    {
      SH_FREE(conn->K);
      conn->K           = NULL;
    }
  if (conn->A != NULL) 
    {
      SH_FREE(conn->A);
      conn->A           = NULL;
    }
  if (conn->M1 != NULL) 
    {
      SH_FREE(conn->M1);
      conn->M1           = NULL;
    }
  if (conn->buf != NULL) 
    {
      SH_FREE(conn->buf);
      conn->buf          = NULL;
    }
  close (conn->fd);
  conn->fd            = -1;
  memset(conn->challenge, '\0', SH_CHALLENGE_SIZE);
  conn->state         = CONN_FREE;
  conn->headcount     = 0;
  conn->bytecount     = 0;
  conn->bytes_to_send = 0;
  conn->bytes_to_get  = 0;
  conn->pass          = 0;
  conn->timer         = 0;
  conn->client_entry  = NULL;

  if (conn->FileName != NULL) 
    {
      SH_FREE(conn->FileName);
      conn->FileName     = NULL;
    }
  conn->FileLength     = 0;
  conn->FileSent       = 0;
  memset(conn->FileType, '\0', 5);

  --server_status.conn_open;
  
  return;
}

/* -- reconfiguration --
 */

/* -- mark all clients as dead
 */
void sh_forward_mark_dead (client_t   * here)
{
  if (here == NULL)
    return;

  sh_forward_mark_dead (here->left);
  sh_forward_mark_dead (here->right);

  here->dead_flag = 1;
  return;
}

/* -- remove a dead client
 */
void sh_forward_remove_client(client_t   * here)
{
  client_t   * child;
  client_t   * next;

  if (here == NULL)
    return;

  if (here->left == NULL || here->right == NULL)
    {
      next = here;
    }
  else
    {
      /* two childs, search next on right side (could search next on
       * left side as well instead)
       */
      next = here->right;        /* cannot be NULL                */
      while (next->left != NULL) /* search leftmost non-NULL node */
	next = next->left;
    }

  if (next->left != NULL)
    child = next->left;
  else
    child = next->right;   /* may be NULL */

  /* replace 'next' by 'child' in the tree
   */
  if (child != NULL)
    child->up = next->up;

  if (next->up == NULL)
    all_clients = child;
  else
    {
      if (next == next->up->left)
	next->up->left  = child;
      else
	next->up->right = child;
    }

  /* and remove 'here'
   */ 
  if (here != next)
    {
      /* replace 'here' by 'next' and free it
       */
      next->left  = here->left;
      next->right = here->right;
      if (next->left  != NULL)
	next->left->up  = next;
      if (next->right != NULL)
	next->right->up = next;
      if (here->up != NULL)
	{
	  if (here == here->up->left)
	    here->up->left  = next;
	  else
	    here->up->right = next;
	}
      else
	all_clients = next;
      free_client(here);
    }
  else
    {
      /* 'here' has already been replaced by 'child', so just free it
       */
      free_client(here);
    }
  return;
}

/* -- clean tree from dead clients
 */
void sh_forward_clean_tree (client_t   * here)
{
  if (here == NULL)
    return;

  sh_forward_clean_tree (here->left);
  sh_forward_clean_tree (here->right);

  if (here->dead_flag == 1)
    sh_forward_remove_client (here);

  return;
}




/* -- send/receive --
 */
void sh_forward_prep_send (sh_conn_t * conn, 
			   char * msg, unsigned long length,
			   char * u, char protocol)
{
  register unsigned long i;

  conn->headcount     = 0;
  conn->bytecount     = 0;
  conn->bytes_to_send = 0;
  conn->bytes_to_get  = 0;

  if (conn->buf != NULL) 
    {
      SH_FREE(conn->buf);
      conn->buf           = NULL;
    }

  put_header (conn->head, protocol, &length, u);


  if (msg == NULL) 
    length = 0;
  
  conn->buf       = SH_ALLOC(length+1);
  for (i = 0; i < length; ++i) 
    conn->buf[i] = msg[i];
  conn->buf[length] = '\0';
  
  conn->state     = CONN_SENDING;
  return;
}


#include <sys/times.h>

#if defined(WITH_EXTERNAL)
#include "sh_extern.h"
#endif

/* Update the client status
 */
static void status_update (sh_conn_t * conn, int status)
{ 
#if defined(WITH_EXTERNAL)
  char msg[2 * SH_MINIBUF + TIM_MAX + 3];
#endif

  if (conn == NULL || conn->client_entry == NULL || 
      status < 0   || status >= CLT_MAX)
    return;
  conn->client_entry->status_now = status;
  conn->client_entry->status_arr[status] = status;
  sl_strlcpy(conn->client_entry->timestamp[status],
	     sh_unix_time(0), 
	     TIM_MAX);

#if defined(WITH_EXTERNAL)
  sprintf(msg, _("%s %s %s"),                           /* known to fit  */
	  conn->peer, 
	  conn->client_entry->timestamp[status],
	  _(clt_stat[status]));
  sh_ext_execute('s', 'r', 'v', msg, 0);
#endif

  return;
}

static time_t time_client_limit = 86400;

int sh_forward_set_time_limit (char * c)
{
  long val;

  val = strtol (c, (char **)NULL, 10);
  if (val <= 0)
    sh_error_handle (SH_ERR_ERR,
		     _("Client limit timer <= NULL: %s"),
		     c);
  val = (val < 0 ? 0 : val);

  time_client_limit = (time_t) val;
  return 0;
}



static int client_time_check(client_t   * top)
{
  int status;

  if ( (top == NULL) || (time_client_limit == (time_t) 0)  )
    return 0;

  /* First, go to the left
   */
  status = client_time_check (top->left);

  if (SL_ISERROR(status))
    return (-1);

  if (top->exit_flag == 0 && top->last_connect != (time_t) 0)
    {
      if ( (time(NULL) - top->last_connect) > time_client_limit)
	{
	  sh_error_handle(SH_ERR_SEVERE,
			  _("ERR_CLIENT -- Time limit (%ld sec) for "\
			    "host %s exceeded."),
			  time_client_limit,
			  top->hostname);
	}
    }

  /* .. then go to the right.
   */
  status = client_time_check (top->right);
  if (SL_ISERROR(status))
    return (-1);

  return 0;
}


  

/************************************************************************
 *
 * Here we check the message received, and decide on the answer to send
 * (if any). The connection is in CONN_PAUSED state, thus we must:
 * (i)   define the proper reaction
 * (ii)  reset to CONN_READING or CONN_WRITING or CONN_FREE
 * (iii) eventually reset the connection entry
 *
 *************************************************************************/
static
void check_protocol(sh_conn_t * conn, int state)
{
  client_t         * this_client; 
  int                treetest;

  char       hash[KEY_LEN+1];
  char     * buffer;
  long       len;
  /* struct tms ttx; */
  UINT32     ticks;
  int        i;
  char     * test;
  char       u[5] = "OOOO";

  SL_TICKET  sfd = -1;
  char      *read_buf;
  char      *send_buf;
  int        bytes;


#ifdef USE_SRP_PROTOCOL
  char     * foo_B;
  char     * foo_Ss;
#endif


  /* seed / re-seed the PRNG if required
   */
  (void) taus_seed();


  /* protocols: 
   * -- (iii)    file transfer
   * -- (ii)     authenticated message transfer
   * -- (i)      SRP key exchange
   */

  /* --------- FILE TRANSFER  -----------
   */
  if ( (conn->head[0] & SH_PROTO_SRP) == 0  &&
       (conn->head[0] & SH_PROTO_BIG) != 0  /* is set */ )
    {

      if (state == SH_DO_READ)        /* finished reading */
	{
	  sh_error_handle(SH_ERR_ALL,
			  _("File transfer - entry."));

	  /* client requests challenge
	   */
	  if (0 == check_request_nerr ((char *) &(conn->head[3]), _("HELO")))
	    {

	      sh_error_handle(SH_ERR_ALL,
			      _("File transfer - HELO (1)."));

	      if (conn->buf == NULL || sl_strlen(conn->buf) <= KEY_LEN)
		{
		  sh_error_handle(SH_ERR_SEVERE,
				  _("BAD_CLIENT -- hostname is NULL"));
		  sh_forward_do_free (conn);
		  return;
		} 

	      /* ----  search the register  -----
	       */
	      this_client = all_clients;
	      while (this_client)
		{
		  treetest = sl_strncmp (&(conn->buf[KEY_LEN]), 
				      this_client->hostname, 
				      sl_strlen(this_client->hostname));
		  if (treetest == 0)
		    break;
		  if (treetest < 0)
		    this_client = this_client->left;
		  else
		    this_client = this_client->right;
		}
	      if (this_client == NULL)
		{
		  sh_error_handle(SH_ERR_SEVERE,
				  _("BAD_CLIENT -- Connection attempt "\
				    "from unregistered host %s"),
				  &(conn->buf[KEY_LEN]));
		  sh_forward_do_free (conn);
		  return;
		} 
	      if (this_client->exit_flag == 1)
		{
		  this_client->session_key_timer = (time_t) 0;
		  this_client->session_key[0]    = '\0';
		  this_client->exit_flag         = 0;    
		}

	      /* ---- force authentication -----
	       */
	      if (this_client->session_key[0] == '\0' ||
		  (time(NULL) - this_client->session_key_timer) 
		  > (time_t) 7200 ) 
		{
		  /* fake an auth request and jump there
		   */
		  conn->head[0]  = (conn->head[0] | SH_PROTO_SRP);
		  conn->head[3]  = 'S';
		  conn->head[4]  = 'A';
		  conn->head[5]  = 'L';
		  conn->head[6]  = 'T';
		  sh_error_handle(SH_ERR_INFO,
				  _("Force authentication "\
				    "for host %s"),
				  &(conn->buf[KEY_LEN]));
		  strcpy(conn->buf, &(conn->buf[KEY_LEN])); /* known to fit  */
		  this_client->session_key[0]    = '\0';
		  this_client->session_key_timer = (time_t) 1;
		  goto servInit;
		}

	      /* check whether hostname is properly signed
	       */ 
	      if (conn->K != NULL) 
		SH_FREE(conn->K); 
	      i = sl_strlen(&(conn->buf[KEY_LEN])) + KEY_LEN + 1;
	      conn->K = SH_ALLOC(i);
	      sl_strlcpy (conn->K, &(conn->buf[KEY_LEN]), i);
	      sl_strlcat (conn->K, this_client->session_key, i);
	      sl_strlcpy (conn->K, 
			  sh_tiger_hash(conn->K, TIGER_DATA, 
					sl_strlen(conn->K)), 
			  KEY_LEN+1);
	      if (0 != sl_strncmp(conn->K, conn->buf, KEY_LEN))
		{
		  sh_error_handle(SH_ERR_SEVERE,
				  _("BAD_CLIENT -- unauthorized "
				    "connection attempt for %s"),
				  &(conn->buf[KEY_LEN]));
		  sh_forward_do_free (conn);
		  return;
		}
	      SH_FREE(conn->K); conn->K = NULL;

	      
	      conn->client_entry = this_client;
	      sl_strlcpy (conn->peer, &(conn->buf[KEY_LEN]), SH_MINIBUF+1);

	      ticks = (UINT32) taus_get (&(skey->rng0[0]), 
					 &(skey->rng1[0]),
					 &(skey->rng2[0]));
	      if (conn->K != NULL) SH_FREE(conn->K);
	      conn->K = SH_ALLOC(KEY_LEN+1);
	      sl_strlcpy (conn->K, 
			  sh_tiger_hash ((char *) &ticks, 
					 TIGER_DATA, sizeof(UINT32)), 
			  KEY_LEN+1);

	      sh_forward_prep_send (conn, conn->K, KEY_LEN+1, _("NSRV"), 
				    SH_PROTO_BIG);
	    }

	  /* client has send a message. check state and message
	   */
	  else if (0 == check_request_nerr((char*)&(conn->head[3]), _("NCLT")) &&
		   conn->client_entry != NULL                           &&
		   sl_strlen(conn->buf) > KEY_LEN                       &&
		   conn->K != NULL)
	    {
	      sh_error_handle(SH_ERR_ALL,
			      _("File transfer - NCLT (3)."));

	      /* get client nonce and compute hash
	       */
	      if (conn->A != NULL)
		SH_FREE(conn->A);
	      conn->A = SH_ALLOC(3*KEY_LEN+1);
	      sl_strlcpy (conn->A, conn->K, KEY_LEN+1); 
	      sl_strlcat(conn->A, conn->buf, 
			 2*KEY_LEN+1);
	      sl_strlcat(conn->A, conn->client_entry->session_key, 
			 3*KEY_LEN+1);
	      sl_strlcpy (conn->K, sh_tiger_hash(conn->A,TIGER_DATA,3*KEY_LEN),
			  KEY_LEN+1);
	      SH_FREE(conn->A); 
	      conn->A = NULL;

	      /* ---- K = H(NSRV, NCLT, session_key) -------
	       */ 

	      if (conn->FileName != NULL)
		SH_FREE(conn->FileName);
	      conn->FileName = NULL;

	      if (0 == sl_strncmp (_("CONF"), &(conn->buf[KEY_LEN]), 4))
		{
		  strcpy(conn->FileType, _("CONF"));
		  conn->FileName = get_client_conf_file(conn->peer, 
							&(conn->FileLength));
		  conn->FileSent = 0;
		}
	      else  if (0 == sl_strncmp (_("DATA"), &(conn->buf[KEY_LEN]), 4))
		{
		  strcpy(conn->FileType, _("DATA"));
		  conn->FileName = get_client_data_file(conn->peer, 
							&(conn->FileLength));
		  conn->FileSent = 0;
		}
	      else
		{
		  /* fprintf(stderr, "FIXME %s\n", conn->buf); */
		  sh_error_handle(SH_ERR_SEVERE,
				  _("ERR_CLIENT %s -- unknown file type "
				    "requested."),
				  conn->peer);
		  status_update (conn, CLT_FAILED);
		  sh_forward_do_free (conn);
		}


	      if (conn != NULL && conn->FileName != NULL)
		{
		  sfd = sl_open_read(conn->FileName, SL_YESPRIV);
		  if (!SL_ISERROR(sfd))
		    {
		      read_buf = SH_ALLOC(TRANS_BYTES);
		      bytes = sl_read (sfd, read_buf, TRANS_BYTES);
		      sl_close(sfd);
		      send_buf = hash_me(conn->K, read_buf, bytes);
		      sh_forward_prep_send (conn, send_buf, bytes+KEY_LEN, 
					    _("FILE"),  SH_PROTO_BIG);
		      conn->FileSent += bytes;
		      SH_FREE(send_buf);
		      SH_FREE(read_buf);
		    }
		}

	      if (conn->FileName == NULL || SL_ISERROR(sfd))
		{
		  sh_error_handle(SH_ERR_SEVERE,
				  _("ERR_CLIENT %s -- %s file not found."),
				  conn->peer, 
				  (conn->FileName == NULL) ? 
				  _("(NULL)") : conn->FileName);
		  status_update (conn, CLT_FAILED);
		  sh_forward_do_free (conn);
		}

	    }
	  
	  else if (0 == check_request_nerr((char*)&(conn->head[3]), _("RECV")) &&
		   conn->client_entry != NULL                           &&
		   conn->K != NULL                                      &&
		   conn->FileName != NULL)
	    {
	      sh_error_handle(SH_ERR_ALL,
			      _("File transfer - RECV (5+)."));

	      if (conn->FileSent == conn->FileLength)
		{
		  send_buf = hash_me(conn->K, conn->peer, 
				     sl_strlen(conn->peer));
		  sh_forward_prep_send (conn, send_buf, 
					sl_strlen(conn->peer)+KEY_LEN, 
					_("EEOT"),  SH_PROTO_BIG);
		}
	      else
		{
		  sfd = sl_open_read(conn->FileName, SL_YESPRIV);
		  if (!SL_ISERROR(sfd))
		    {
		      read_buf = SH_ALLOC(TRANS_BYTES);
		      sl_seek (sfd, (off_t) conn->FileSent);
		      bytes = sl_read (sfd, read_buf, TRANS_BYTES);
		      sl_close(sfd);

		      send_buf = hash_me(conn->K, read_buf, bytes);
		      sh_forward_prep_send (conn, send_buf, bytes+KEY_LEN, 
					    _("FILE"),  SH_PROTO_BIG);
		      conn->FileSent += bytes;
		      SH_FREE(send_buf);
		      SH_FREE(read_buf);
		    }
		  else
		    {
		      sh_error_handle(SH_ERR_SEVERE,
				      _("ERR_CLIENT %s -- file not found."),
				      conn->peer);
		      status_update (conn, CLT_FAILED);
		      sh_forward_do_free (conn);
		    }
		}
	    }


	  else if (0 == check_request_nerr((char*)&(conn->head[3]), _("EOTE")) &&
		   conn->client_entry != NULL)
	    {
	      sh_error_handle(SH_ERR_ALL,
			      _("File transfer - EOTE (7)."));
	      sh_error_handle(SH_ERR_INFO,
			      _("File transfer finished."));

	      status_update (conn, CLT_FILE);
	      conn->client_entry->last_connect = time (NULL);
	      sh_forward_do_free (conn);
	    }
	    

	  /* client does something unexpected
	   */
	  else  /* ---- ??? ----- */
	    {
	      sh_error_handle(SH_ERR_SEVERE,
			      _("ERR_CLIENT %s -- file transfer (pass %d), "\
				"request %c%03o%c%03o%c%03o%c%03o invalid."),
			      conn->peer, conn->pass, 
			      '\\', conn->head[3], '\\',conn->head[4],
			      '\\', conn->head[5], '\\',conn->head[6]);
	      status_update (conn, CLT_FAILED);
	      sh_forward_do_free (conn);
	    }
	}

      else if (state == SH_DO_WRITE)  /* finished writing */
	{
	  sh_error_handle(SH_ERR_ALL,
			  _("File transfer - (wait)."));

	  /* challenge is sent, now wait for message from client
	   */
	  conn->headcount     = 0;
	  conn->bytecount     = 0;
	  conn->bytes_to_send = 0;
	  conn->bytes_to_get  = 0;
	  if (conn->buf != NULL) 
	    {
	      SH_FREE(conn->buf);
	      conn->buf           = NULL;
	    }
	  conn->state     = CONN_READING;
	}
      return;
    }

  /* ---------  message exchange  -----------
   */
  if ((conn->head[0] & SH_PROTO_SRP) == 0  && 
      (conn->head[0] & SH_PROTO_MSG) != 0  /* is set */ )
    {

      if (state == SH_DO_READ)        /* finished reading */
	{

	  sh_error_handle(SH_ERR_ALL,
			  _("Msg transfer - entry."));

	  /* client requests challenge
	   */
	  if (0 == check_request_nerr ((char*)&(conn->head[3]), _("HELO")))
	    {

	      sh_error_handle(SH_ERR_ALL,
			      _("Msg transfer - HELO (1)."));

	      if (conn->buf == NULL || sl_strlen(conn->buf) <= KEY_LEN )
		{
		  sh_error_handle(SH_ERR_SEVERE,
				  _("BAD_CLIENT -- hostname is NULL"));
		  sh_forward_do_free (conn);
		  return;
		} 

	      /* ----  search the register  -----
	       */
	      this_client = all_clients;
	      while (this_client)
		{
		  treetest = sl_strncmp (&(conn->buf[KEY_LEN]), 
				      this_client->hostname, 
				      sl_strlen(this_client->hostname));
		  if (treetest == 0)
		    break;
		  if (treetest < 0)
		    this_client = this_client->left;
		  else
		    this_client = this_client->right;
		}
	      if (this_client == NULL)
		{
		  sh_error_handle(SH_ERR_ERR,
				  _("Connection attempt "\
				    "from unregistered host %s"),
				  &(conn->buf[KEY_LEN]));
		  sh_forward_do_free (conn);
		  return;
		}
 
	      if (this_client->exit_flag == 1)
		{
		  this_client->session_key_timer = (time_t) 0;
		  this_client->session_key[0]    = '\0';
		  this_client->exit_flag         = 0;    
		}


	      /* ---- force authentication -----
	       */
	      if (this_client->session_key[0] == '\0' || 
		  (time(NULL)-this_client->session_key_timer) > (time_t)7200
		  ) 
		{
		  /* fake an auth request and jump there
		   */
		  conn->head[0]  = (conn->head[0] | SH_PROTO_SRP);
		  conn->head[3]  = 'S';
		  conn->head[4]  = 'A';
		  conn->head[5]  = 'L';
		  conn->head[6]  = 'T';
		  sh_error_handle(SH_ERR_INFO,
				  _("Force authentication "\
				    "for host %s"),
				  &(conn->buf[KEY_LEN]));
		  strcpy(conn->buf, &(conn->buf[KEY_LEN])); /* known to fit  */
		  this_client->session_key[0]    = '\0';
		  this_client->session_key_timer = (time_t) 1;

		  goto servInit;
		}
	      
	      /* check whether hostname is properly signed
	       */ 
	      if (conn->K != NULL) 
		SH_FREE(conn->K);
	      i = sl_strlen(&(conn->buf[KEY_LEN])) + KEY_LEN + 1;
	      conn->K = SH_ALLOC(i);
	      sl_strlcpy (conn->K, &(conn->buf[KEY_LEN]), i);
	      sl_strlcat (conn->K, this_client->session_key, i);

	      sl_strlcpy (conn->K, 
			  sh_tiger_hash(conn->K, TIGER_DATA, 
					sl_strlen(conn->K)), 
			  KEY_LEN+1);
	      if (0 != sl_strncmp(conn->K, conn->buf, KEY_LEN))
		{
		  sh_error_handle(SH_ERR_SEVERE,
				  _("BAD_CLIENT -- unauthorized "
				    "connection attempt for %s"),
				  &(conn->buf[KEY_LEN]));
		  sh_forward_do_free (conn);
		  return;
		}
	      SH_FREE(conn->K); conn->K = NULL;


	      conn->client_entry = this_client;
	      sl_strlcpy (conn->peer, &(conn->buf[KEY_LEN]), SH_MINIBUF+1);

	      ticks = (UINT32) taus_get (&(skey->rng0[0]), 
					 &(skey->rng1[0]),
					 &(skey->rng2[0]));
	      test = (char *) &ticks;
	      sh_util_cpylong (conn->challenge, test, 4);
	      conn->challenge[4] = '\0';
	      for (i = 0; i < 4; ++i)
		if (conn->challenge[i] == '\0')
		  conn->challenge[i] = 0x01;
	      sh_forward_prep_send (conn, conn->challenge, 5, _("TALK"), 
				    SH_PROTO_MSG);
	    }

	  /* Client has send a message. Check whether we are in proper
	   * state, and verify message.
	   */
	  else if (0 == check_request_nerr((char*)&(conn->head[3]), _("MESG")) &&
		   conn->client_entry != NULL                           &&
		   conn->client_entry->session_key[0] != '\0'           &&
		   (len = sl_strlen(conn->buf) - KEY_LEN) > 0              &&
		   sl_strlen(conn->challenge) == 4)
	    {
	      sh_error_handle(SH_ERR_ALL,
			      _("Msg transfer - MESG (3)."));

	      /* get hash from message end, truncate message
	       */
	      sl_strlcpy(hash, &(conn->buf[len]), KEY_LEN+1);
	      conn->buf[len] = '\0';
	      
	      /* verify hash
	       */
	      buffer = sh_util_strconcat(conn->buf, conn->challenge,
					 conn->client_entry->session_key,
					 NULL);

	      if (0 != sl_strncmp(hash, 
			       sh_tiger_hash (buffer, 
					      TIGER_DATA, 
					      sl_strlen(buffer)),
			       KEY_LEN))
		{
		  status_update (conn, CLT_FAILED);
		  sh_error_handle(SH_ERR_ERR,
				  _("Bogus connection from %s"),
				  conn->peer);
		}
	      else
		{
		  conn->client_entry->last_connect = time (NULL);

		  if (NULL != sl_strstr(conn->buf,      _("EXIT")))
		    {
		      conn->client_entry->exit_flag = 1;
		      status_update (conn, CLT_EXITED);
		    }
		  else if (NULL != sl_strstr(conn->buf, _("PANIC")))
		    {
		      status_update (conn, CLT_PANIC);
		    }
		  else if (NULL != sl_strstr(conn->buf, _("POLICY")))
		      status_update (conn, CLT_POLICY);
		  else if (NULL != sl_strstr(conn->buf, _("STARTUP")))
		    {
		      if (NULL != sl_strstr(conn->buf, _("BADSIG")) ||
			  NULL != sl_strstr(conn->buf, _("ERRSIG")))
			{ 
			  status_update (conn, CLT_POLICY);
			}
		    }
		  else
		    status_update (conn, CLT_MSG);

		  sh_error_handle(SH_ERR_INET,
				  _("<%s> %s"),
				  sh_strip_domain (conn->peer),
				  conn->buf);
		}
	      memset(buffer, '\0', sl_strlen(buffer));
	      SH_FREE(buffer);

	      /* SEND CONFIRMATION
	       */
	      buffer = sh_util_strconcat(conn->buf,
					 conn->client_entry->session_key,
					 conn->challenge,
					 NULL);

	      sl_strlcpy(hash, 
			 sh_tiger_hash (buffer, 
					TIGER_DATA, 
					sl_strlen(buffer)),
			 KEY_LEN+1);

	      sh_forward_prep_send (conn, hash, KEY_LEN, _("CONF"), 
				    SH_PROTO_MSG|SH_PROTO_END);

	      memset(buffer, '\0', sl_strlen(buffer));
	      SH_FREE(buffer);
	      
	      /* sh_forward_do_free (conn); */
	    }
	  
	  /* client does something unexpected
	   */
	  else  /* ---- ??? ----- */
	    {
	      sh_error_handle(SH_ERR_SEVERE,
			      _("ERR_CLIENT %s -- message transfer (pass %d),"\
				" request %c%03o%c%03o%c%03o%c%03o invalid."),
			      conn->peer, conn->pass, 
			      '\\', conn->head[3], '\\',conn->head[4],
			      '\\',conn->head[5], '\\',conn->head[6]);
	      sh_forward_do_free (conn);
	    }
	}
      else if (state == SH_DO_WRITE)  /* finished writing */
	{
	  if (0 != (conn->head[0] & SH_PROTO_END))
	    {
	      sh_error_handle(SH_ERR_INFO,
			      _("Msg transfer finished."));
	      sh_forward_do_free (conn);
	      return;
	    }

	  sh_error_handle(SH_ERR_ALL,
			  _("Msg transfer - (wait)."));

	  /* challenge is sent, now wait for message from client
	   */
	  conn->headcount     = 0;
	  conn->bytecount     = 0;
	  conn->bytes_to_send = 0;
	  conn->bytes_to_get  = 0;
	  if (conn->buf != NULL) 
	    {
	      SH_FREE(conn->buf);
	      conn->buf           = NULL;
	    }
	  conn->state     = CONN_READING;
	}
      return;
    }

  /* ---------  authentication  -----------
   */

  /* entry point for jump from message forward if session key must
   * be re-initialized
   */	 
 servInit:

  if ( (conn->head[0] & SH_PROTO_SRP) != 0   /* is set */ )
    {

#ifndef USE_SRP_PROTOCOL

      if (state == SH_DO_READ)        /* finished reading */
	{
	  sh_error_handle(SH_ERR_ALL,
			  _("Authentication - entry."));

	  /* first pass -- client request salt  
	   */
	  if (conn->pass    == 1) 
	    {

	      sh_error_handle(SH_ERR_ALL,
			      _("Authentication - SALT (1)."));

	      if (conn->buf == NULL || sl_strlen(conn->buf) == 0)
		{
		  sh_error_handle(SH_ERR_SEVERE,
				  _("BAD_CLIENT -- hostname is NULL"));
		  sh_forward_do_free (conn);
		  return;
		} 
		  

	      /* search the register
	       */
	      this_client = all_clients;
	      while (this_client)
		{
		  treetest = sl_strncmp (conn->buf, 
				      this_client->hostname, 
				      sl_strlen(this_client->hostname));
		  if (treetest == 0)
		    break;
		  if (treetest < 0)
		    this_client = this_client->left;
		  else
		    this_client = this_client->right;
		}
	      if (this_client == NULL)
		{
		  sh_error_handle(SH_ERR_SEVERE,
				  _("BAD_CLIENT %s -- unregistered host"),
				  conn->buf);
		  sh_forward_do_free (conn);
		  return;
		} 
	      if (this_client->exit_flag == 1)
		{
		  this_client->session_key_timer = (time_t) 0;
		  this_client->session_key[0]    = '\0';
		  this_client->exit_flag         = 0;    
		}



	      conn->client_entry = this_client;
	      sl_strlcpy (conn->peer, conn->buf, SH_MINIBUF+1);

	      if (0 != check_request_s((char*)&(conn->head[3]), _("SALT"),conn->peer))
		{
		  sh_error_handle(SH_ERR_SEVERE,
				  _("ERR_CLIENT %s -- auth (1) invalid "
				    "request"), conn->peer);
		  status_update (conn, CLT_FAILED);
		  sh_forward_do_free (conn);
		  return;
		} 

	      /* 
	       * create server nounce v
	       */
	      ticks = (UINT32) taus_get (&(skey->rng0[0]), 
					 &(skey->rng1[0]),
					 &(skey->rng2[0]));
	      
	      if (conn->A != NULL)
		SH_FREE(conn->A);
	      conn->A = SH_ALLOC(KEY_LEN+1);

	      sl_strlcpy(conn->A, 
			 sh_tiger_hash((char *) &ticks, 
				       TIGER_DATA, sizeof(UINT32)),
			 KEY_LEN+1);
	      u[0] = 'I'; u[1] = 'N'; u[2] = 'I'; u[3] = 'T'; u[4] = '\0';

	      if (conn->M1 != NULL)
		SH_FREE(conn->M1);
	      conn->M1 = SH_ALLOC(2*KEY_LEN+1);

	      /* compute hash key H(v(server), P)v(server)
	       */
	      sh_passwd (conn->A, conn->client_entry->verifier, NULL,conn->M1);

	      sl_strlcat(conn->M1, conn->A, 2*KEY_LEN+1);


	      /* --- send H(v(server), P)v(server) ----
	       */
      	      sh_forward_prep_send (conn, 
				    conn->M1, 
				    sl_strlen(conn->M1), 
				    u, 
				    (conn->head[0]|SH_PROTO_SRP));

	      SH_FREE(conn->M1); 
	      conn->M1 = NULL;
	    }

	  /* client -- third pass
	   * Message is H(H(u,v),P)u
	   *
	   * A := v, verifier := H(password), 
	   */
	  else if (conn->pass    == 3                   && 
		   conn->client_entry != NULL)
	    {

	      sh_error_handle(SH_ERR_ALL,
			      _("Authentication - PASS (3)."));
				  
	      if (0 != check_request_s((char*)&(conn->head[3]), _("PASS"), 
				       conn->peer)                    ||
		  sl_strlen(conn->buf) <= KEY_LEN                        ||
		  conn->A == NULL)
		{
		  sh_error_handle(SH_ERR_SEVERE,
				  _("ERR_CLIENT %s -- auth (3) invalid "
				    "connection state"),
				  conn->peer);
		  status_update (conn, CLT_FAILED);
		  sh_forward_do_free (conn);
		  return;
		} 

	      /* store random nonce u from client
	       */
	      if (conn->K != NULL)
		SH_FREE(conn->K);
	      conn->K = SH_ALLOC(KEY_LEN+1);
	      sl_strlcpy(conn->K, &(conn->buf[KEY_LEN]), KEY_LEN+1);

	      /* verify random nonce u from client
	       */
	      if (conn->M1 != NULL)
		SH_FREE(conn->M1);
	      conn->M1 = sh_util_strconcat(conn->K, conn->A, NULL);
#ifdef SRP_DEBUG
	      fprintf(stderr, "K  <%s>\n", conn->K);
	      fprintf(stderr, "A  <%s>\n", conn->A);
	      fprintf(stderr, "M  <%s>\n", conn->M1);
#endif
	      sl_strlcpy(hash, sh_tiger_hash (conn->M1, 
					      TIGER_DATA, 
					      sl_strlen(conn->M1)), KEY_LEN+1); 
	      sh_passwd (hash, conn->client_entry->verifier, NULL, conn->M1);

#ifdef SRP_DEBUG
	      fprintf(stderr, "H  <%s>\n", hash);
	      fprintf(stderr, "P  <%s>\n", conn->M1);
#endif

	      if ( 0 != sl_strncmp(conn->M1, conn->buf, KEY_LEN))
		{
		  sh_error_handle(SH_ERR_SEVERE,
				  _("ERR_CLIENT %s -- auth (3) "\
				    "invalid data"),
				  conn->peer);
		  status_update (conn, CLT_FAILED);
		  sh_forward_do_free (conn);
		  return;
		} 
		  

	      /* ---- compute hash key H(v, P, u) ----
	       */
	      
	      sh_passwd (conn->A, conn->client_entry->verifier, conn->K,
			 conn->M1);

	      sl_strlcpy(conn->client_entry->session_key, 
			 conn->M1, KEY_LEN+1);

	      if (conn->K  != NULL) SH_FREE (conn->K);
	      conn->K  = NULL;
	      if (conn->A  != NULL) SH_FREE (conn->A);
	      conn->A  = NULL;
	      if (conn->M1 != NULL) SH_FREE (conn->M1);
	      conn->M1 = NULL;

	      if (conn->client_entry->status_now == CLT_STARTED &&
		  conn->client_entry->session_key_timer > (time_t) 1)
		{
		  status_update (conn, CLT_ILLEGAL);
		  
		  sh_error_handle(SH_ERR_SEVERE,
				  _("ERR_CLIENT %s -- restart without "\
				    "prior exit."),
				  conn->peer);
		}
	      else if (conn->client_entry->session_key_timer == (time_t) 0)
		{
		  sh_error_handle(SH_ERR_SEVERE,
				  _("NEW_CLIENT %s"),
				  conn->peer);
		  status_update (conn, CLT_STARTED);
		}

	      conn->client_entry->session_key_timer = time (NULL);
	      conn->client_entry->last_connect = time (NULL);

	      /* put in read state
	       */
	      sh_forward_prep_send (conn, 
				    _("AUTH"),
				    5, 
				    _("AUTH"), 
				    (conn->head[0]|SH_PROTO_SRP));

	    }
	  else
	    {
	      sh_error_handle(SH_ERR_SEVERE,
			      _("ERR_CLIENT %s -- authentication (pass %d), "\
				"request %c%03o%c%03o%c%03o%c%03o invalid."),
			      conn->peer, conn->pass, 
			      '\\', conn->head[3], '\\', conn->head[4],
			      '\\', conn->head[5], '\\', conn->head[6]);
	      sh_forward_do_free (conn);
	    }

	}


#else
      /* use SRP                    */
      /* #include "srp_server.frag" */


      if (state == SH_DO_READ)        /* finished reading */
	{

	  sh_error_handle(SH_ERR_ALL,
			  _("Authentication - entry."));

	  /* first pass -- client request salt  
	   */
	  if (conn->pass    == 1)
	    {
	      sh_error_handle(SH_ERR_ALL,
			      _("Authentication - SALT (1)."));
	      if (conn->buf == NULL)
		{
		  sh_error_handle(SH_ERR_SEVERE,
				  _("BAD_CLIENT -- hostname is NULL"));
		  sh_forward_do_free (conn);
		  return;
		} 

	      /* search the register
	       */
	      this_client = all_clients;
	      while (this_client)
		{
		  treetest = sl_strncmp (conn->buf, 
					 this_client->hostname, 
					 sl_strlen(this_client->hostname));
		  if (treetest == 0)
		    break;
		  if (treetest < 0)
		    this_client = this_client->left;
		  else
		    this_client = this_client->right;
		}
	      if (this_client == NULL)
		{
		  sh_error_handle(SH_ERR_SEVERE,
				  _("BAD_CLIENT %s -- unregistered host"),
				  conn->buf);
		  sh_forward_do_free (conn);
		  return;
		} 
	      if (this_client->exit_flag == 1)
		{
		  this_client->session_key_timer = (time_t) 0;
		  this_client->session_key[0]    = '\0';
		  this_client->exit_flag         = 0;    
		}

	      conn->client_entry = this_client;
	      sl_strlcpy (conn->peer, conn->buf, SH_MINIBUF+1);

	      if (0 != check_request_s((char*)&(conn->head[3]), _("SALT"),
				       conn->peer))
		{
		  sh_error_handle(SH_ERR_SEVERE,
				  _("ERR_CLIENT %s -- auth (1) invalid "
				    "connection state"),
				  conn->peer);
		  status_update (conn, CLT_FAILED);
		  sh_forward_do_free (conn);
		  return;
		} 
		  

	      u[0] = 'I'; u[1] = 'N'; u[2] = 'I'; u[3] = 'T'; u[4] = '\0';
	      
	      sh_forward_prep_send (conn, 
				    conn->client_entry->salt, 
				    sl_strlen(conn->client_entry->salt), 
				    u, 
				    (conn->head[0]|SH_PROTO_SRP));
	    }

	  /* client has sent A -- third pass
	   */
	  else if (conn->pass == 3                    && 
		   conn->client_entry != NULL)
	    {

	      sh_error_handle(SH_ERR_ALL,
			      _("Authentication - PC01 (3)."));

	      if (0 != check_request_s((char*)&(conn->head[3]),_("PC01"),conn->peer)||
		  conn->buf == NULL
		  )
		{
		  sh_error_handle(SH_ERR_SEVERE,
				  _("ERR_CLIENT %s -- invalid connection state"),
				  conn->peer);
		  status_update (conn, CLT_FAILED);
		  sh_forward_do_free (conn);
		  return;
		} 

	      if (0 != sh_srp_init())
		{
		  status_update (conn, CLT_FAILED);
		  sh_error_handle(SH_ERR_SEVERE,
				  _("ERR_CLIENT %s -- Bignum library "\
				    "internal error (init)."),
				  conn->peer);
		  sh_error_handle(SH_ERR_SEVERE,
				  _("ERR_CLIENT %s -- Bignum library "\
				    "internal error (init)."),
				  conn->peer);
		  sh_forward_do_free (conn);
		  return;
		}
		  

	      /* check A, only send B if correct 
	       */
	      if ( sl_strlen(conn->buf) < SH_BUFSIZE && 
		   0 == sh_srp_check_zero (conn->buf) )
		{
		  len = sl_strlen(conn->buf)+1;

		  if (conn->A != NULL)
		    SH_FREE(conn->A);
		  conn->A = SH_ALLOC(len);
		  sl_strlcpy (conn->A, conn->buf, len);
		  
		  /* 
		   * compute B 
		   */
		  if (0 != sh_srp_make_a ())     /* b        random number */
		    {
		      status_update (conn, CLT_FAILED);
		      
		      sh_error_handle(SH_ERR_SEVERE,
				      _("ERR_CLIENT %s -- Bignum library "\
					"internal error (b)."),
				      conn->peer);
		      sh_srp_exit();
		      sh_forward_do_free (conn);
		      return;
		    }
             
		  foo_B = sh_srp_B               /* B = v + g^b            */
		    (conn->client_entry->verifier);

		  if (foo_B == NULL)
		    {
		      status_update (conn, CLT_FAILED);
		      
		      sh_error_handle(SH_ERR_SEVERE,
				      _("ERR_CLIENT %s -- Bignum library "\
					"internal error (B)."),
				      conn->peer);
		      sh_srp_exit();
		      sh_forward_do_free (conn);
		      return;
		    }

#ifdef SRP_DEBUG
		  fprintf(stderr, "DD Server: A: <%s>\n", conn->A);
		  fprintf(stderr, "DD Server: B: <%s>\n", foo_B);
#endif
		  /* 
		   * create nonce u 
		   */
		  ticks = (UINT32) taus_get (&(skey->rng0[0]), 
					     &(skey->rng1[0]),
					     &(skey->rng2[0])); 
		  test = (char *) &ticks;
		  sh_util_cpylong (u, test, 4);  /* u        nounce        */
		  u[4] = '\0';
		  sl_strlcpy(conn->challenge, 
			     sh_tiger_hash(u, TIGER_DATA, 4),
			     SH_CHALLENGE_SIZE);

#ifdef SRP_DEBUG
		  fprintf(stderr, "DD Server: u: <%s>\n", u);
		  fprintf(stderr, "DD Server: U: <%s>\n", conn->challenge);
#endif

		  /* 
		   * compute the session key K and M1 = Hash(A,B,K)
		   */
		  foo_Ss = sh_srp_S_s (conn->challenge, 
				       conn->A, 
				       conn->client_entry->verifier);
		  if (foo_Ss == NULL)
		    {
		      status_update (conn, CLT_FAILED);
		      
		      sh_error_handle(SH_ERR_SEVERE,
				      _("ERR_CLIENT %s -- Bignum library "\
					"internal error (Ss)."),
				      conn->peer);
		      sh_srp_exit();
		      sh_forward_do_free (conn);
		      return;
		    }

		  if (conn->K != NULL)
		    SH_FREE(conn->K);
		  conn->K = SH_ALLOC(KEY_LEN+1);
		  sl_strlcpy(conn->K, 
			     sh_tiger_hash(foo_Ss, TIGER_DATA, 
					   sl_strlen(foo_Ss)),
			     KEY_LEN+1);

		  if (conn->M1 != NULL)
		    SH_FREE(conn->M1);
		  conn->M1 = SH_ALLOC(KEY_LEN+1);
		  sl_strlcpy(conn->M1, 
			  sh_srp_M (conn->A, foo_B, conn->K),
			  KEY_LEN+1);

#ifdef SRP_DEBUG
		  fprintf(stderr, "DD Server:Ss: <%s>\n", foo_Ss);
		  fprintf(stderr, "DD Server: K: <%s>\n", conn->K);
		  fprintf(stderr, "DD Server:M1: <%s>\n", conn->M1);
#endif

		  /*
		   * send B
		   */
		  sh_forward_prep_send (conn, 
					foo_B,
					sl_strlen(foo_B)+1,
					u,
					(conn->head[0]|SH_PROTO_SRP));
		  if (foo_Ss != NULL) 
		    SH_FREE(foo_Ss);
		  if (foo_B  != NULL)
		    SH_FREE(foo_B);
		}
	      else
		{
		  status_update (conn, CLT_FAILED);

		  sh_error_handle(SH_ERR_SEVERE,
				  _("ERR_CLIENT %s -- SRP protocol: A == 0"),
				  conn->peer);
		  sh_forward_do_free (conn);
		}

	      sh_srp_exit();
	    }
		  
	  /* client has sent M1 -- fifth pass
	   */
	  else if (conn->pass    == 5           && 
		   conn->client_entry != NULL) 
	    {
	      sh_error_handle(SH_ERR_ALL,
			      _("Authentication - PC02 (5)."));

	      /* check that the state is valid
	       */
	      if (0 != check_request_s((char*)&(conn->head[3]), _("PC02"),
				       conn->peer)                   ||
		  conn->A == NULL || conn->K == NULL || conn->M1 == NULL)
		{
		  sh_error_handle(SH_ERR_SEVERE,
				  _("ERR_CLIENT %s -- invalid connection state"),
				  conn->peer);
		  status_update (conn, CLT_FAILED);
		  sh_forward_do_free (conn);
		  return;
		} 

	      /* ------ verify M1 = H(A,  B, K) -------
	       * -----    send M2 = H(A, M1, K) -------
	       */
	      if (conn->buf != NULL && 
		  sl_strncmp(conn->buf, conn->M1, KEY_LEN) == 0)
		{
		  /*
		   * send M2
		   */
		  sh_forward_prep_send (conn, 
					sh_srp_M (conn->A, conn->M1, conn->K),
					KEY_LEN+1,
					_("PARP"),
					(conn->head[0]|SH_PROTO_SRP));

		  if (conn->A  != NULL) SH_FREE(conn->A);  conn->A  = NULL;
		  if (conn->M1 != NULL) SH_FREE(conn->M1); conn->M1 = NULL;
		  sl_strlcpy(conn->client_entry->session_key, 
			     conn->K, KEY_LEN+1);
		  if (conn->K  != NULL) SH_FREE(conn->K);  conn->K  = NULL;

		  conn->client_entry->last_connect = time (NULL);
		  
		  if (conn->client_entry->status_now == CLT_STARTED &&
		      conn->client_entry->session_key_timer > (time_t) 1)
		    {
		      status_update (conn, CLT_ILLEGAL);

		      sh_error_handle(SH_ERR_SEVERE,
				      _("ERR_CLIENT %s -- restart without "\
					"prior exit."),
				      conn->peer);
		    }
		  else if (conn->client_entry->session_key_timer == (time_t) 0)
		    {
		      sh_error_handle(SH_ERR_SEVERE,
				      _("NEW_CLIENT %s"),
				      conn->peer);
		      status_update (conn, CLT_STARTED);
		    }
		  conn->client_entry->session_key_timer = time (NULL);

		}
	      else
		{
		  status_update (conn, CLT_FAILED);

		  sh_error_handle(SH_ERR_SEVERE, 
				  _("ERR_CLIENT %s -- could not verify "\
				    "session key"),
				  conn->peer);
		  sh_forward_do_free (conn);
		} 
	    }

	  else
	    {
	      sh_error_handle(SH_ERR_SEVERE,
			      _("ERR_CLIENT %s -- authentication (pass %d), "\
				"request %c%03o%c%03o%c%03o%c%03o invalid."),
			      conn->peer, conn->pass, 
			      '\\', conn->head[3], '\\', conn->head[4],
			      '\\', conn->head[5], '\\', conn->head[6]);
	      sh_forward_do_free (conn);
	    }
	}

#endif

      else if (state == SH_DO_WRITE)  /* finished writing */
	{
	  sh_error_handle(SH_ERR_ALL,
			  _("Authentication -- (wait)."));

	  conn->headcount     = 0;
	  conn->bytecount     = 0;
	  conn->bytes_to_send = 0;
	  conn->bytes_to_get  = 0;
	  if (conn->buf != NULL) 
	    {
	      SH_FREE(conn->buf);
	      conn->buf           = NULL;
	    }
	  conn->state     = CONN_READING;
	}
    }
}


/* read from to the client
 */
int sh_forward_do_read (sh_conn_t * conn)
{
  unsigned long   byteread;     /* bytes read         */

  if (conn->state == CONN_SENDING)
    {
      sh_error_handle(SH_ERR_SEVERE, 
		      _("ERR_CLIENT %s -- out of sync"),
		      conn->peer);
      return (-1);
    }

  if (conn->headcount < SH_HEADER_SIZE) 
    {
      conn->bytes_to_get = SH_HEADER_SIZE - conn->headcount;
      byteread           = read (conn->fd, &(conn->head[conn->headcount]),
				    conn->bytes_to_get);
      if (byteread > 0 || errno == EINTR) 
	{
	  if (byteread > 0) 
	    conn->headcount += byteread;
	  if (conn->headcount == SH_HEADER_SIZE)
	    {
		conn->bytes_to_get = 
		  (256 * (int)conn->head[1] + (int)conn->head[2]);
	    }
	}
      else
	{
	  sh_error_handle(SH_ERR_SEVERE, 
			  _("ERR_CLIENT %s -- connection reset by peer"),
			  conn->peer);
	  sh_forward_do_free ( conn );
	  return (-1);
	}
      return (0);
    }


  /* limit message size
   */
  conn->bytes_to_get = 
    (conn->bytes_to_get > 4*SH_BUFSIZE) ? (4*SH_BUFSIZE) : conn->bytes_to_get;


  if (conn->headcount == SH_HEADER_SIZE && conn->bytes_to_get > 0)
    {
      if (conn->buf != NULL)
	SH_FREE (conn->buf);

      conn->buf       = SH_ALLOC (conn->bytes_to_get + 1);
      conn->bytecount = 0;

      byteread           = read (conn->fd, &(conn->buf[conn->bytecount]),
				 conn->bytes_to_get - conn->bytecount);
      if (byteread > 0 || errno == EINTR) 
	{
	  if (byteread > 0) 
	    conn->bytecount    += byteread;
	  if (conn->bytecount == conn->bytes_to_get) 
	    {
	      ++conn->pass;
	      /* always terminate with NULL - we might use sl_strcmp()
	       */
	      conn->buf[conn->bytecount] = '\0';
	      conn->state                = CONN_PAUSE;
	      /* ------  HERE CALL check_protocol(conn) -------  */
	      check_protocol(conn, SH_DO_READ);
	    }
	}
      else
	{
	  sh_error_handle(SH_ERR_SEVERE, 
			  _("ERR_CLIENT %s -- connection reset by peer"),
			  conn->peer);
	  sh_forward_do_free ( conn );
	  return (-1);
	}
    }

  else if (conn->headcount == SH_HEADER_SIZE && conn->bytes_to_get == 0)
    {
      /* ------  HERE CALL check_protocol(conn) -------  */
      check_protocol(conn, SH_DO_READ);
    }
      
  return (0);
}

/* send to the client
 */
int sh_forward_do_write (sh_conn_t * conn)
{
  unsigned long   bytesent;     /* bytes read         */

  /* ---- consistency check ------
   */
  if (conn->state == CONN_READING)
    {
      sh_error_handle(SH_ERR_SEVERE, 
		      _("ERR_CLIENT %s -- out of sync"),
		      conn->peer);
      return (-1);
    }
      

  /* ---- send the header ------
   */
  if (conn->headcount < SH_HEADER_SIZE) 
    {
      conn->bytes_to_send = SH_HEADER_SIZE - conn->headcount;
      bytesent            = write (conn->fd, 
				   &(conn->head[conn->headcount]), 
				   conn->bytes_to_send);
      if (bytesent >= 0 || errno == EINTR) 
	{
	  if (bytesent > 0) 
	    conn->headcount += bytesent;
	  if (conn->headcount == SH_HEADER_SIZE) 
	    {
	      conn->bytes_to_send = 
		(256 * (int)conn->head[1] + (int)conn->head[2]);
	    }
	}
      else 
	{
	  sh_error_handle(SH_ERR_SEVERE, 
			  _("ERR_CLIENT %s -- connection reset by peer"),
			  conn->peer);
	  sh_forward_do_free ( conn );
	  return (-1);
	}
      return (0);
    }


  /* ---- send the body ------
   */

  if (conn->headcount == SH_HEADER_SIZE && conn->bytes_to_send > 0 &&
      conn->buf != NULL)
    {
      bytesent           = write (conn->fd, &(conn->buf[conn->bytecount]), 
				     conn->bytes_to_send - conn->bytecount);
      if (bytesent >= 0 || errno == EINTR) 
	{
	  if (bytesent > 0) 
	    conn->bytecount    += bytesent;
	  if (conn->bytecount == conn->bytes_to_send) 
	    {
	      ++conn->pass;
	      conn->state         = CONN_PAUSE;
	      /* ------  HERE CALL check_protocol(conn) -------  */
	      check_protocol(conn, SH_DO_WRITE);
	    }
	}
      else
	{
	  sh_error_handle(SH_ERR_SEVERE, 
			  _("ERR_CLIENT %s -- connection reset by peer"),
			  conn->peer);
	  sh_forward_do_free ( conn );
	  return (-1);
	}
    }
      
  else if (conn->headcount == SH_HEADER_SIZE && conn->bytes_to_send == 0)
    {
      /* ------  HERE CALL check_protocol(conn) -------  */
      check_protocol(conn, SH_DO_WRITE);
    }

  return (0);
}

/* accept a connection from a client
 */ 
int sh_forward_accept (int sock, sh_conn_t * newconn)
{
  int                rc;
  struct sockaddr_in addr;
#ifdef HOST_IS_AIX
  size_t             addrlen = sizeof(addr);
#else
  int                addrlen = sizeof(addr);
#endif
  int                errflag;

  if ( (rc = EINTR_RETRY(accept(sock, (struct sockaddr *) &addr, &addrlen))) 
       >= 0)
    {

      /* prepare for usage of connection
       */
      (void) EINTR_RETRY(fcntl( rc, F_SETFD, 1 ));
      newconn->fd           = rc;
      newconn->state        = CONN_READING;
      newconn->timer        = (unsigned long) time (NULL);

      sh_error_handle(SH_ERR_INFO,
		      _("New connection, using file descriptor %d"),
		      rc);
      return (0);
    }
  else
    {
      errflag = errno;
      sh_error_handle(SH_ERR_ERR,
		      _("Connection error: accept: %s"),
		      sh_error_message(errflag));

      newconn->fd    = -1;
      newconn->state = CONN_FREE;
      return (-1);
    }

}

extern char sh_sig_msg[64];  /* defined in sh_unix.c */

/* print error
 */
void sh_forward_printerr(char * str, int errnum)
{
  sh_error_handle(SH_ERR_ERR,
		  _("Server error: %s: %s"),
		  str,
		  sh_error_message(errnum));
  return;
}

static    sh_conn_t        * conns = NULL;

#define  TIME_OUT_DEF 300

static  int  maxconn = 0;  /* maximum number of simultaneous connections */

/*****************************************
 *
 * This is the server main loop.
 *
 * The server is set up for listening, and
 * and starts a select() loop.
 *
 *****************************************/
void sh_receive()
{

  struct sockaddr_in addr;
  int addrlen      = sizeof(addr);
  
  int                sock;
  int                flag = 1;  /* non-zero to enable an option */
  
  sh_conn_t        * cx;
  fd_set             readset;
  fd_set             writeset;
  struct timeval     tv;
  int                num_sel;
  int                errnum;
  int                nowconn;
  int                status;
  int                high_fd;
  register int       i;
  unsigned long      time_now;
  unsigned long      time_last = 0;
  unsigned long      time_out = TIME_OUT_DEF;  
  
  unsigned long told;
  unsigned long tcurrent;

  struct  sigaction  new_act;
  struct  sigaction  old_act;
  
  /* ignore SIGPIPE (instead get EPIPE if connection is closed)
   *      --- we have called sh_unix_init() already ---
   */
  new_act.sa_handler = SIG_IGN;
  sigaction (SIGPIPE, &new_act, &old_act);
  
  /* create the socket, bind() it and listen()
   */
  if ((sock = socket(AF_INET, SOCK_STREAM, 0)) < 0 )
    {
      errnum = errno; 
      sh_forward_printerr (_("socket"), errnum);
      exit(EXIT_FAILURE);
    }
  (void) EINTR_RETRY(fcntl( sock, F_SETFD, 1 ));
  
  if ( setsockopt(sock, SOL_SOCKET, SO_REUSEADDR,
		  (void *) &flag, sizeof(flag)) < 0 )
    {
      errnum = errno;
      sh_forward_printerr (_("setsockopt"), errnum);
      exit(EXIT_FAILURE);
    }
  
  addr.sin_family = AF_INET;
  addr.sin_port = htons(SH_DEFAULT_PORT);
  addr.sin_addr.s_addr = INADDR_ANY;
  
  if ( bind(sock, (struct sockaddr *) &addr, addrlen) < 0) 
    {
      errnum = errno;
      sh_forward_printerr (_("bind"), errnum);
      exit(EXIT_FAILURE);
    }
  
  if ( EINTR_RETRY(fcntl( sock, F_SETFL, O_NDELAY )) < 0 )
    {
      errnum = errno;
      sh_forward_printerr (_("fcntl"), errnum);
      exit(EXIT_FAILURE);
    }
  
  if ( listen(sock, 5) < 0)
    {
      errnum = errno;
      sh_forward_printerr (_("listen"), errnum);
      exit(EXIT_FAILURE);
    }
  
  /* ****************************************************************
   *
   * This is a non-forking server. We use select() on the listen()
   * socket to watch for new connections. For new connections, accept()
   * will return a new socket that is put in the read/write filesets.
   * Data about active connections are kept in the 'conns' table. 
   *
   ******************************************************************/
  
  /* The table to hold info on sockets.
   * We reserve 6 file descriptors for misc. use.
   * The POSIX lower limit on open files seems to be eight. 
   */
  maxconn = get_open_max() - 6;
  maxconn = (FD_SETSIZE < maxconn) ? FD_SETSIZE : maxconn;

  conns   = SH_ALLOC (sizeof(sh_conn_t) * maxconn);

  sh_error_handle(SH_ERR_STAMP,
		  _("Server: listen fd %d, simultaneous connections: %d."),
		  sock, (maxconn-1));

  /* timer
   */
  tcurrent                   = (unsigned long) time (NULL);
  told                       = tcurrent;
  
  for (i = 1; i < maxconn; ++i)
    {
      conns[i].buf         = NULL;
      conns[i].K           = NULL;
      conns[i].A           = NULL;
      conns[i].M1          = NULL;
      conns[i].FileName    = NULL;
      sh_forward_do_free ( &conns[i]);
    }
  
  /* status init
   */
  server_status.conn_open  = 0;
  server_status.conn_total = 0;
  server_status.conn_max   = maxconn-1;
  server_status.start      = time (NULL);
  server_status.last       = (time_t) 0;

  nowconn    = 1;
  tv.tv_sec  = 5;
  tv.tv_usec = 0;
  
  /* conns[0] is the listen() socket. Always in read mode.
   */
  conns[0].fd    = sock;
  conns[0].state = CONN_READING;
  
  high_fd = sock;
  
  sh_html_write(all_clients);
  
  /* This is the select() loop.
   */
  while (1 == 1)
    {

    if (sig_raised != 0)
      {

	if (sig_config_read_again == 1)
	  {
	    sh_error_handle(SH_ERR_SEVERE,
			    _("--- RECONFIGURATION ---"));
	    /* delete the name server cache
	     */
	    delete_cache();
#if defined(WITH_EXTERNAL)
	    /* delete list of external tasks
	     */
	    (void) sh_ext_cleanup();
#endif
	    /* - mark all clients dead
	     * - read configuration file
	     * - remove clients still dead
	     */
	    sh_forward_mark_dead (all_clients);
	    (void) sh_readconf_read ();
	    for (i = 1; i < maxconn; ++i)
	      if (conns[i].state != CONN_FREE   && 
		  conns[i].client_entry != NULL &&
		  conns[i].client_entry->dead_flag == 1)
		sh_forward_do_free ( &conns[i]);
	    sh_forward_clean_tree (all_clients);
	    sig_config_read_again = 0;
	    sig_raised            = 0;
	  }

	if (sig_fresh_trail == 1) /* SIGIOT */
	  {
	    /* Logfile access 
	     */
	    sh_error_only_stderr (GOOD);
	    sh_unix_unlock();
	    sh.isfirst     = 1;
	    sh.firstaccess = 1;
	    sleep(3);
	    sh_error_only_stderr (BAD);
	    sig_fresh_trail       = 0;
	    sig_raised            = 0;
	  }

	if (sig_terminate == 1 && nowconn < 2)  /* SIGQUIT */
	  {
	    strncpy (sh_sig_msg, _("on signal SIGQUIT"), 20);
	    exit (EXIT_SUCCESS);
	  }

	if (sig_debug_on == 1)  /* SIGUSR1 */
	  {
	    sh_error_dbg_on();
	    sig_debug_on  = 0;
	    sig_raised    = 0;
	  }

	if (sig_debug_off == 1)  /* SIGUSR2 */
	  {
	    sh_error_dbg_off();
	    sig_debug_off = 0;
	    sig_raised    = 0;
	  }
      }

      
      /* Recompute the descriptor set. select() modifies it,
       * thus we update it using the info from the connection table.
       * Also recompute the number of open connections.
       */
      FD_ZERO( &readset );
      FD_ZERO( &writeset );
      FD_SET(conns[0].fd, &readset );
      high_fd   = conns[0].fd;
      time_now  = (unsigned long) time (NULL);
      nowconn   = 1;
      
      for (i = 1; i < maxconn; ++i)
	{
	  /* eliminate timed out connections
	   */
	  if (conns[i].state != CONN_FREE) 
	    {
	      if (time_now-conns[i].timer > time_out)
		{
		  sh_error_handle(SH_ERR_ERR,
				  _("Connection to %s timed out"),
				  conns[i].peer);
		  sh_forward_do_free ( &conns[i]);
		}
	      else
		++nowconn;
	    }
	  
	  /*****************************************************
	   * 
	   * Drop connections at random if more than 4/5 full.
	  if (nowconn > (4*(maxconn/5)) && conns[i].state != CONN_FREE)
	    { 
	      if ( (5.0 * rand() / (RAND_MAX+1.0)) > 4)
		{ 
		  sh_error_handle(SH_ERR_ERR,
				  _("Connection to %s dropped"),
				  conns[i].peer);
		  sh_forward_do_free ( &conns[i]);
		}
	    }
	    *
	    *******************************************************/
	  
	  if       (conns[i].state   == CONN_READING)
	    { 
	      FD_SET(conns[i].fd, &readset);
	      high_fd = (high_fd < conns[i].fd ? conns[i].fd : high_fd);
	    }
	  else if  (conns[i].state   == CONN_SENDING)
	    {
	      FD_SET(conns[i].fd, &writeset);
	      high_fd = (high_fd < conns[i].fd ? conns[i].fd : high_fd);
	    }
	}

      /* Exponentially reduce timeout limit if more than 1/2 full.
       */
      if (nowconn > (maxconn/2))
	time_out = ( (time_out/2) > 1) ? (time_out/2) : 1;
      else
	time_out = TIME_OUT_DEF;
      
      
      
      /* Do the select().
       */
      num_sel = select(high_fd+1, &readset, &writeset, NULL, &tv);
      errnum  = errno;
      
      /* reset timeout - modified by select() on some systems
       */
      tv.tv_sec  = 5;
      tv.tv_usec = 0;
      

      if ( (time_now - time_last) > 2L)
	{
	  time_last = time_now;
	  if (sh_html_write(all_clients) < 0)
	    sh_error_handle(SH_ERR_ERR,
			    _("Error writing html report."));
	}
      
      
      /* Error handling.
       */
      if ( num_sel < 0 )        /* some error             */
	{
	  if (sig_raised == 1)
	    {
	      sig_raised = 2;
	      continue;
	    }
	  if ( errnum == EINTR)
	    continue;	  /* try again              */
	  sh_error_handle(SH_ERR_FATAL, 
			  _("PANIC -- fatal server error - select() %s"),
			  sh_error_message(errnum));
	  exit( EXIT_FAILURE );
	}
      
      /* select() timeout handling.
       */
      if ( num_sel == 0 )       /* timeout - no connection */ 
	{
	  if (sh_html_write(all_clients) < 0)
	    sh_error_handle(SH_ERR_ERR,
			    _("Error writing html report."));
	  continue;
	}

      /* New connection.
       */
      if ( FD_ISSET(conns[0].fd , &readset )) /* a new connection   */
	{
	  --num_sel;
	  status = 0;
	  if (nowconn < maxconn && sig_terminate == 0)
	    {
	      i = 1;
	      while (i < maxconn)
		{
		  if (conns[i].state == CONN_FREE)
		    {
		      status = sh_forward_accept (conns[0].fd, &conns[i]);
		      if (status == 0)
			{
			  high_fd = 
			    (high_fd > conns[i].fd ? high_fd : conns[i].fd);
			  ++server_status.conn_open;
			  ++server_status.conn_total;
			  server_status.last = time (NULL);
			}
		      break;
		    }
		  ++i;
		}
	    }
	  if (status == 0) continue;
	}
      
      /* Check for pending read/write on the rest of the sockets.
       */
      for ( i = 0; num_sel > 0 && i < maxconn; ++i )
	{
	  cx = &conns[i];
	  if ( cx->state == CONN_READING &&
	       FD_ISSET( cx->fd, &readset ) )
	    {
	      --num_sel;
	      sh_forward_do_read ( cx );
	    }
	  else if ( cx->state == CONN_SENDING &&
		    FD_ISSET( cx->fd, &writeset ) )
	    {
	      --num_sel;
	      sh_forward_do_write ( cx );
	    }
	}
      
      /* log the timestamp
       */
      if ((tcurrent - told) > (unsigned int) sh.looptime )
	{
	  told = tcurrent;
#ifdef MEM_DEBUG
	  sh_mem_check();
#else
	  sh_error_handle (SH_ERR_STAMP,
			   _("---- TIMESTAMP ----"));
#endif
	}
      tcurrent = (unsigned long) time (NULL);


      /* check for time limit exceeded
       */
      client_time_check(all_clients);
      
      /* seed / re-seed the PRNG if required
       */
      (void) taus_seed();
      
    }
}

void  free_client_tree (client_t * inptr)
{
  if (inptr == NULL)
    return;

  /* free left and right tree
   */
  free_client_tree (inptr->left);
  free_client_tree (inptr->right);

  /* free this node
   */
  if (inptr->hostname) SH_FREE(inptr->hostname);
  if (inptr->salt)     SH_FREE(inptr->salt);
  if (inptr->verifier) SH_FREE(inptr->verifier);
  inptr->left  = NULL;
  inptr->right = NULL;
  inptr->up    = NULL;
  SH_FREE(inptr);
  inptr        = NULL;

  return;
}

void sh_forward_free_all ()
{
  register int i;
  
  if (conns != NULL)
    for (i = 1; i < maxconn; ++i)
      {
	sh_forward_do_free ( &conns[i]);
      }


  free_client_tree (all_clients);

  if (conns != NULL)
    SH_FREE (conns);

  return;
}
/* #ifdef SH_WITH_SERVER */
#endif


  
